#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import constants
from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy import sql

from neutron.db.models import l3 as l3_models
from neutron.db import models_v2


class NetworkDNSDomain(model_base.BASEV2):
    network_id = sa.Column(sa.String(36),
                           sa.ForeignKey('networks.id', ondelete="CASCADE"),
                           primary_key=True)
    dns_domain = sa.Column(sa.String(255),
                           nullable=False)

    # Add a relationship to the Network model in order to instruct
    # SQLAlchemy to eagerly load this association
    network = orm.relationship(models_v2.Network,
                               load_on_pending=True,
                               backref=orm.backref("dns_domain",
                                                   lazy='joined',
                                                   uselist=False,
                                                   cascade='delete'))
    revises_on_change = ('network', )


class FloatingIPDNS(model_base.BASEV2):

    __tablename__ = 'floatingipdnses'

    floatingip_id = sa.Column(sa.String(36),
                              sa.ForeignKey('floatingips.id',
                                            ondelete="CASCADE"),
                              primary_key=True)
    dns_name = sa.Column(sa.String(255),
                         nullable=False)
    dns_domain = sa.Column(sa.String(255),
                           nullable=False)
    published_dns_name = sa.Column(sa.String(255),
                                   nullable=False)
    published_dns_domain = sa.Column(sa.String(255),
                                     nullable=False)

    # Add a relationship to the FloatingIP model in order to instruct
    # SQLAlchemy to eagerly load this association
    floatingip = orm.relationship(l3_models.FloatingIP,
                                  load_on_pending=True,
                                  backref=orm.backref("dns",
                                                      lazy='joined',
                                                      uselist=False,
                                                      cascade='delete'))
    revises_on_change = ('floatingip', )


class PortDNS(model_base.BASEV2):

    __tablename__ = 'portdnses'

    port_id = sa.Column(sa.String(36),
                        sa.ForeignKey('ports.id',
                                      ondelete="CASCADE"),
                        primary_key=True)
    current_dns_name = sa.Column(sa.String(255),
                                 nullable=False)
    current_dns_domain = sa.Column(sa.String(255),
                                   nullable=False)
    previous_dns_name = sa.Column(sa.String(255),
                                  nullable=False)
    previous_dns_domain = sa.Column(sa.String(255),
                                    nullable=False)
    dns_name = sa.Column(sa.String(255), nullable=False)
    dns_domain = sa.Column(sa.String(constants.FQDN_FIELD_SIZE),
                           nullable=False,
                           server_default='')
    # Add a relationship to the Port model in order to instruct
    # SQLAlchemy to eagerly load this association
    port = orm.relationship(models_v2.Port,
                            load_on_pending=True,
                            backref=orm.backref("dns",
                                                lazy='joined',
                                                uselist=False,
                                                cascade='delete'))
    revises_on_change = ('port', )


class SubnetDNSPublishFixedIP(model_base.BASEV2):
    __tablename__ = "subnet_dns_publish_fixed_ips"

    subnet_id = sa.Column(sa.String(constants.UUID_FIELD_SIZE),
                          sa.ForeignKey('subnets.id', ondelete="CASCADE"),
                          primary_key=True)
    dns_publish_fixed_ip = sa.Column(sa.Boolean(),
                                     nullable=False,
                                     server_default=sql.false())

    # Add a relationship to the Subnet model in order to instruct
    # SQLAlchemy to eagerly load this association
    subnet = orm.relationship(models_v2.Subnet,
                              load_on_pending=True,
                              backref=orm.backref("dns_publish_fixed_ip",
                                                  lazy='joined',
                                                  uselist=False,
                                                  cascade='delete'))
    revises_on_change = ('subnet', )



# Copyright 2016 Hewlett Packard Enterprise Development Company, LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import testtools

from neutron_lib.api.definitions import trunk as trunk_api
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from neutron_lib.plugins.ml2 import api
from neutron_lib.plugins import utils as plugin_utils
from neutron_lib.services.trunk import constants
from oslo_utils import uuidutils

from neutron.services.trunk import drivers
from neutron.services.trunk import exceptions as trunk_exc
from neutron.services.trunk import plugin as trunk_plugin
from neutron.services.trunk import rules
from neutron.services.trunk import utils as trunk_utils
from neutron.tests import base
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron.tests.unit.services.trunk import fakes


class SubPortsValidatorTestCase(base.BaseTestCase):

    def setUp(self):
        super(SubPortsValidatorTestCase, self).setUp()
        self.segmentation_types = {
            constants.SEGMENTATION_TYPE_VLAN: plugin_utils.is_valid_vlan_tag}
        self.context = mock.ANY

        mock.patch.object(rules.SubPortsValidator, '_get_port_mtu',
                          return_value=None).start()
        mock.patch.object(rules.SubPortsValidator, '_prepare_subports',
                          return_value=None).start()

    def test_validate_subport_subport_and_trunk_shared_port_id(self):
        shared_id = uuidutils.generate_uuid()
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'port_id': shared_id,
              'segmentation_type': 'vlan',
              'segmentation_id': 2}],
            shared_id)
        self.assertRaises(trunk_exc.ParentPortInUse,
                          validator.validate, self.context)

    def test_validate_subport_invalid_vlan_id(self):
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'port_id': uuidutils.generate_uuid(),
              'segmentation_type': 'vlan',
              'segmentation_id': 5000}])
        self.assertRaises(n_exc.InvalidInput,
                          validator.validate,
                          self.context)

    def test_validate_subport_vlan_id_not_an_int(self):
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'port_id': uuidutils.generate_uuid(),
              'segmentation_type': 'vlan',
              'segmentation_id': 'IamNotAnumber'}])
        self.assertRaises(n_exc.InvalidInput,
                          validator.validate,
                          self.context)

    def test_validate_subport_valid_vlan_id_as_string(self):
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'port_id': uuidutils.generate_uuid(),
              'segmentation_type': 'vlan',
              'segmentation_id': '2'}])
        with mock.patch.object(rules.TrunkPortValidator, 'validate') as f:
            validator.validate(self.context)
            f.assert_called_once_with(self.context, parent_port=False)

    def test_validate_subport_subport_invalid_segmentation_type(self):
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'port_id': uuidutils.generate_uuid(),
              'segmentation_type': 'fake',
              'segmentation_id': 100}])
        self.assertRaises(n_exc.InvalidInput,
                          validator.validate,
                          self.context)

    def test_validate_subport_missing_segmentation_type(self):
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'port_id': uuidutils.generate_uuid(),
              'segmentation_id': 100}])
        self.assertRaises(n_exc.InvalidInput,
                          validator.validate,
                          self.context)

    def test_validate_subport_missing_segmentation_id(self):
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'port_id': uuidutils.generate_uuid(),
              'segmentation_type': 'fake'}])
        self.assertRaises(n_exc.InvalidInput,
                          validator.validate,
                          self.context)

    def test_validate_subport_missing_port_id(self):
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'segmentation_type': 'fake',
            'segmentation_id': 100}])
        self.assertRaises(n_exc.InvalidInput,
                          validator.validate,
                          self.context, basic_validation=True)


class SubPortsValidatorPrepareTestCase(base.BaseTestCase):

    def setUp(self):
        super(SubPortsValidatorPrepareTestCase, self).setUp()
        self.segmentation_types = {
            constants.SEGMENTATION_TYPE_VLAN: plugin_utils.is_valid_vlan_tag}
        self.context = mock.ANY

        mock.patch.object(rules.SubPortsValidator, '_get_port_mtu',
                          return_value=None).start()

    def test__prepare_subports_raise_no_provider_ext(self):
        validator = rules.SubPortsValidator(
            self.segmentation_types,
            [{'port_id': uuidutils.generate_uuid(),
              'segmentation_type': 'inherit'}])
        self.assertRaises(n_exc.InvalidInput,
                          validator._prepare_subports,
                          self.context)


class SubPortsValidatorMtuSanityTestCase(test_plugin.Ml2PluginV2TestCase):

    def setUp(self):
        super(SubPortsValidatorMtuSanityTestCase, self).setUp()
        self.segmentation_types = {
            constants.SEGMENTATION_TYPE_VLAN: plugin_utils.is_valid_vlan_tag}

    def test_validate_subport_mtu_same_as_trunk(self):
        self._test_validate_subport_trunk_mtu(1500, 1500)

    def test_validate_subport_mtu_smaller_than_trunks(self):
        self._test_validate_subport_trunk_mtu(500, 1500)

    def test_validate_subport_mtu_greater_than_trunks(self):
        self._test_validate_subport_trunk_mtu(1500, 500)

    def test_validate_subport_mtu_unset_trunks_set(self):
        self._test_validate_subport_trunk_mtu(None, 500)

    def test_validate_subport_mtu_set_trunks_unset(self):
        self._test_validate_subport_trunk_mtu(500, None)

    def test_validate_subport_mtu_set_trunks_net_exception(self):
        self._test_validate_subport_trunk_mtu(1500, 'exc')

    def _test_validate_subport_trunk_mtu(
            self, subport_net_mtu, trunk_net_mtu):
        plugin = directory.get_plugin()
        orig_get_network = plugin.get_network
        orig_get_networks = plugin.get_networks

        def get_networks_adjust_mtu(*args, **kwargs):
            res = orig_get_networks(*args, **kwargs)
            res[0][api.MTU] = subport_net_mtu
            return res

        def get_network_adjust_mtu(*args, **kwargs):
            res = orig_get_network(*args, **kwargs)
            if res['name'] == 'net_trunk':
                if trunk_net_mtu == 'exc':
                    raise n_exc.NetworkNotFound(net_id='net-id')
                res[api.MTU] = trunk_net_mtu
            elif res['name'] == 'net_subport':
                res[api.MTU] = subport_net_mtu
            return res

        with self.network('net_trunk') as trunk_net,\
            self.subnet(network=trunk_net) as trunk_subnet,\
            self.port(subnet=trunk_subnet) as trunk_port,\
            self.network('net_subport') as subport_net,\
            self.subnet(network=subport_net) as subport_subnet,\
            self.port(subnet=subport_subnet) as subport,\
            mock.patch.object(plugin, "get_network",
                              side_effect=get_network_adjust_mtu),\
            mock.patch.object(plugin, "get_networks",
                              side_effect=get_networks_adjust_mtu):
            trunk = {'port_id': trunk_port['port']['id'],
                     'tenant_id': 'test_tenant',
                     'sub_ports': [{'port_id': subport['port']['id'],
                                    'segmentation_type': 'vlan',
                                    'segmentation_id': 2}]}

            validator = rules.SubPortsValidator(
                self.segmentation_types, trunk['sub_ports'], trunk['port_id'])

            if subport_net_mtu is None or trunk_net_mtu is None:
                validator.validate(self.context)
            elif subport_net_mtu == 'exc' or trunk_net_mtu == 'exc':
                validator.validate(self.context)
            elif subport_net_mtu <= trunk_net_mtu:
                validator.validate(self.context)
            else:
                self.assertRaises(trunk_exc.SubPortMtuGreaterThanTrunkPortMtu,
                                  validator.validate, self.context)


class TrunkPortValidatorTestCase(test_plugin.Ml2PluginV2TestCase):

    def setUp(self):
        super(TrunkPortValidatorTestCase, self).setUp()
        self.drivers_patch = mock.patch.object(drivers, 'register').start()
        self.compat_patch = mock.patch.object(
            trunk_plugin.TrunkPlugin, 'check_compatibility').start()
        self.trunk_plugin = trunk_plugin.TrunkPlugin()
        self.trunk_plugin.add_segmentation_type(
            constants.SEGMENTATION_TYPE_VLAN,
            plugin_utils.is_valid_vlan_tag)

    def test_validate_port_parent_in_use_by_trunk(self):
        with self.port() as trunk_parent:
            trunk = {'port_id': trunk_parent['port']['id'],
                     'project_id': 'test_tenant',
                     'sub_ports': []}
            self.trunk_plugin.create_trunk(
                self.context, {trunk_api.ALIAS: trunk})
            validator = rules.TrunkPortValidator(trunk_parent['port']['id'])
            self.assertRaises(trunk_exc.ParentPortInUse,
                              validator.validate,
                              self.context)

    def test_validate_port_id_in_use_by_unrelated_trunk(self):
        with self.port() as trunk_parent,\
                 self.port() as subport:
            trunk = {'port_id': trunk_parent['port']['id'],
                     'project_id': 'test_tenant',
                     'sub_ports': [{'port_id': subport['port']['id'],
                                    'segmentation_type': 'vlan',
                                    'segmentation_id': 2}]}
            self.trunk_plugin.create_trunk(
                self.context, {trunk_api.ALIAS: trunk})
            validator = rules.TrunkPortValidator(subport['port']['id'])
            self.assertRaises(trunk_exc.TrunkPortInUse,
                              validator.validate,
                              self.context)

    def test_validate_port_has_binding_host(self):
        with self.port() as port:
            core_plugin = directory.get_plugin()
            port['port']['binding:host_id'] = 'host'
            core_plugin.update_port(self.context, port['port']['id'], port)
            validator = rules.TrunkPortValidator(port['port']['id'])
            self.assertTrue(validator.is_bound(self.context))

    def test_validate_for_subport_calls_check(self):
        with self.port() as port:
            validator = rules.TrunkPortValidator(port['port']['id'])
            with mock.patch.object(validator, "check_not_in_use") as f:
                validator.validate(self.context, parent_port=False)
                f.assert_called_once_with(self.context)

    def test_validate_port_cannot_be_trunked_raises(self):
        with self.port() as port, \
             mock.patch.object(rules.TrunkPortValidator,
                               "can_be_trunked_or_untrunked",
                               return_value=False), \
             testtools.ExpectedException(trunk_exc.ParentPortInUse):
            validator = rules.TrunkPortValidator(port['port']['id'])
            validator.validate(self.context)

    def test_can_be_trunked_or_untrunked_returns_false(self):
        # need to trigger a driver registration
        fakes.FakeDriverCanTrunkBoundPort.create()
        self.trunk_plugin = trunk_plugin.TrunkPlugin()
        directory.add_plugin('trunk', self.trunk_plugin)
        with self.port() as port:
            core_plugin = directory.get_plugin()
            port['port']['binding:host_id'] = 'host'
            core_plugin.update_port(self.context, port['port']['id'], port)
            validator = rules.TrunkPortValidator(port['port']['id'])
            # port cannot be trunked because of binding mismatch
            self.assertFalse(
                validator.can_be_trunked_or_untrunked(self.context))

    def test_can_be_trunked_or_untrunked_returns_true(self):
        # need to trigger a driver registration
        fakes.FakeDriverCanTrunkBoundPort.create()
        self.trunk_plugin = trunk_plugin.TrunkPlugin()
        directory.add_plugin('trunk', self.trunk_plugin)
        with self.port() as port, \
                mock.patch.object(trunk_utils, "is_driver_compatible",
                                  return_value=True) as g:
            core_plugin = directory.get_plugin()
            port['port']['binding:host_id'] = 'host'
            core_plugin.update_port(self.context, port['port']['id'], port)
            validator = rules.TrunkPortValidator(port['port']['id'])
            self.assertTrue(
                validator.can_be_trunked_or_untrunked(self.context))
            self.assertTrue(g.call_count)

    def test_can_be_trunked_or_untrunked_unbound_port(self):
        with self.port() as port:
            validator = rules.TrunkPortValidator(port['port']['id'])
            self.assertTrue(
                validator.can_be_trunked_or_untrunked(self.context))

    def test_can_be_trunked_or_untrunked_raises_conflict(self):
        d1 = fakes.FakeDriver.create()
        d2 = fakes.FakeDriverWithAgent.create()
        self.trunk_plugin = trunk_plugin.TrunkPlugin()
        directory.add_plugin('trunk', self.trunk_plugin)
        self.trunk_plugin._drivers = [d1, d2]
        with self.port() as port, \
                mock.patch.object(trunk_utils, "is_driver_compatible",
                                  return_value=True):
            core_plugin = directory.get_plugin()
            port['port']['binding:host_id'] = 'host'
            core_plugin.update_port(self.context, port['port']['id'], port)
            validator = rules.TrunkPortValidator(port['port']['id'])
            self.assertRaises(
                trunk_exc.TrunkPluginDriverConflict,
                validator.can_be_trunked_or_untrunked, self.context)

    def test_check_not_in_use_pass(self):
        with self.port() as port:
            validator = rules.TrunkPortValidator(port['port']['id'])
            self.assertIsNone(validator.check_not_in_use(
                self.context))

    def test_check_not_in_use_raises(self):
        with self.port() as port:
            core_plugin = directory.get_plugin()
            port['port']['device_id'] = 'foo_device_id'
            core_plugin.update_port(self.context, port['port']['id'], port)
            validator = rules.TrunkPortValidator(port['port']['id'])
            self.assertRaises(n_exc.PortInUse,
                              validator.check_not_in_use, self.context)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


# These tables are in the neutron database, but their models have moved
# to separate repositories. We skip the migration checks for these tables.

VPNAAS_TABLES = ['vpnservices', 'ipsecpolicies', 'ipsecpeercidrs',
                 'ipsec_site_connections', 'cisco_csr_identifier_map',
                 'ikepolicies']

# Neutron-lbaas is retired, but we need to keep this for the models until
# we decide to remove the tables.
LBAAS_TABLES = ['vips', 'sessionpersistences', 'pools', 'healthmonitors',
                'poolstatisticss', 'members', 'poolloadbalanceragentbindings',
                'poolmonitorassociations']

FWAAS_TABLES = ['firewall_rules', 'firewalls', 'firewall_policies']

# Arista ML2 driver Models moved to openstack/networking-arista
REPO_ARISTA_TABLES = [
    'arista_provisioned_nets',
    'arista_provisioned_vms',
    'arista_provisioned_tenants',
]

# BGP models in openstack/neutron-dynamic-routing
REPO_NEUTRON_DYNAMIC_ROUTING_TABLES = [
    'bgp_speakers',
    'bgp_peers',
    'bgp_speaker_network_bindings',
    'bgp_speaker_peer_bindings',
    'bgp_speaker_dragent_bindings',
]

# Models moved to openstack/networking-cisco
REPO_CISCO_TABLES = [
    'cisco_ml2_apic_contracts',
    'cisco_ml2_apic_names',
    'cisco_ml2_apic_host_links',
    'cisco_ml2_n1kv_policy_profiles',
    'cisco_ml2_n1kv_network_profiles',
    'cisco_ml2_n1kv_port_bindings',
    'cisco_ml2_n1kv_network_bindings',
    'cisco_ml2_n1kv_vxlan_allocations',
    'cisco_ml2_n1kv_vlan_allocations',
    'cisco_ml2_n1kv_profile_bindings',
    'cisco_ml2_nexusport_bindings',
    'cisco_ml2_nexus_nve',
    'ml2_nexus_vxlan_allocations',
    'ml2_nexus_vxlan_mcast_groups',
    'ml2_ucsm_port_profiles',
    'cisco_hosting_devices',
    'cisco_port_mappings',
    'cisco_router_mappings',
]

# VMware-NSX models moved to openstack/vmware-nsx
REPO_VMWARE_TABLES = [
    'tz_network_bindings',
    'neutron_nsx_network_mappings',
    'neutron_nsx_security_group_mappings',
    'neutron_nsx_port_mappings',
    'neutron_nsx_router_mappings',
    'multi_provider_networks',
    'networkconnections',
    'networkgatewaydevicereferences',
    'networkgatewaydevices',
    'networkgateways',
    'maclearningstates',
    'qosqueues',
    'portqueuemappings',
    'networkqueuemappings',
    'lsn_port',
    'lsn',
    'nsxv_router_bindings',
    'nsxv_edge_vnic_bindings',
    'nsxv_edge_dhcp_static_bindings',
    'nsxv_internal_networks',
    'nsxv_internal_edges',
    'nsxv_security_group_section_mappings',
    'nsxv_rule_mappings',
    'nsxv_port_vnic_mappings',
    'nsxv_router_ext_attributes',
    'nsxv_tz_network_bindings',
    'nsxv_port_index_mappings',
    'nsxv_firewall_rule_bindings',
    'nsxv_spoofguard_policy_network_mappings',
    'nsxv_vdr_dhcp_bindings',
    'vcns_router_bindings',
]

# Brocade models are in openstack/networking-brocade
REPO_BROCADE_TABLES = [
    'brocadenetworks',
    'brocadeports',
    'ml2_brocadenetworks',
    'ml2_brocadeports',
]

# BigSwitch models are in openstack/networking-bigswitch
REPO_BIGSWITCH_TABLES = [
    'consistencyhashes',
    'routerrules',
    'nexthops',
]

# Nuage models are in github.com/nuagenetworks/nuage-openstack-neutron
REPO_NUAGE_TABLES = [
    'nuage_net_partitions',
    'nuage_net_partition_router_mapping',
    'nuage_provider_net_bindings',
    'nuage_subnet_l2dom_mapping',
]

TABLES = (FWAAS_TABLES + LBAAS_TABLES + VPNAAS_TABLES +
          REPO_ARISTA_TABLES +
          REPO_NEUTRON_DYNAMIC_ROUTING_TABLES +
          REPO_CISCO_TABLES +
          REPO_VMWARE_TABLES +
          REPO_BROCADE_TABLES +
          REPO_BIGSWITCH_TABLES +
          REPO_NUAGE_TABLES)



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import time
from unittest import mock

from neutron_lib import context
from neutron_lib.db import api as db_api
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy.orm import exc

from neutron.db.models import ovn as ovn_models
from neutron.db import ovn_hash_ring_db
from neutron.tests.unit import testlib_api

HASH_RING_TEST_GROUP = 'test_group'


class TestHashRing(testlib_api.SqlTestCaseLight):

    def setUp(self):
        super(TestHashRing, self).setUp()
        self.admin_ctx = context.get_admin_context()
        self.addCleanup(self._delete_objs)

    def _delete_objs(self):
        with db_api.CONTEXT_WRITER.using(self.admin_ctx):
            self.admin_ctx.session.query(
                ovn_models.OVNRevisionNumbers).delete()

    def _get_node_row(self, node_uuid):
        try:
            with db_api.CONTEXT_WRITER.using(self.admin_ctx):
                node = self.admin_ctx.session.query(
                    ovn_models.OVNHashRing).filter_by(
                    node_uuid=node_uuid).one()
            # When a record is created, the difference between "created_at" and
            # "updated_at" should be tiny, just some microseconds.
            if (node.updated_at - node.created_at).total_seconds() < 0.01:
                node.updated_at = node.created_at
            # Ignore miliseconds
            node.created_at = node.created_at.replace(microsecond=0)
            node.updated_at = node.updated_at.replace(microsecond=0)
            return node
        except exc.NoResultFound:
            return

    def _add_nodes_and_assert_exists(self, count=1,
                                     group_name=HASH_RING_TEST_GROUP):
        nodes = []
        for i in range(count):
            node_uuid = ovn_hash_ring_db.add_node(self.admin_ctx, group_name)
            self.assertIsNotNone(self._get_node_row(node_uuid))
            nodes.append(node_uuid)
        return nodes

    def test_add_node(self):
        self._add_nodes_and_assert_exists()

    def test_remove_nodes_from_host(self):
        nodes = self._add_nodes_and_assert_exists(count=3)

        # Add another node from a different host
        with mock.patch.object(ovn_hash_ring_db, 'CONF') as mock_conf:
            mock_conf.host = 'another-host-' + uuidutils.generate_uuid()
            another_host_node = self._add_nodes_and_assert_exists()[0]

        ovn_hash_ring_db.remove_nodes_from_host(self.admin_ctx,
                                                HASH_RING_TEST_GROUP)
        # Assert that all nodes from that host have been removed
        for n in nodes:
            self.assertIsNone(self._get_node_row(n))

        # Assert that the node from another host wasn't removed
        self.assertIsNotNone(self._get_node_row(another_host_node))

    def test_touch_nodes_from_host(self):
        nodes = self._add_nodes_and_assert_exists(count=3)

        # Add another node from a different host
        with mock.patch.object(ovn_hash_ring_db, 'CONF') as mock_conf:
            mock_conf.host = 'another-host-' + uuidutils.generate_uuid()
            another_host_node = self._add_nodes_and_assert_exists()[0]

        # Assert that updated_at isn't updated yet
        for node in nodes:
            node_db = self._get_node_row(node)
            self.assertEqual(node_db.created_at, node_db.updated_at)

        # Assert the same for the node from another host
        node_db = self._get_node_row(another_host_node)
        self.assertEqual(node_db.created_at, node_db.updated_at)

        # Touch the nodes from our host
        time.sleep(1)
        ovn_hash_ring_db.touch_nodes_from_host(self.admin_ctx,
                                               HASH_RING_TEST_GROUP)

        # Assert that updated_at is now updated
        for node in nodes:
            node_db = self._get_node_row(node)
            self.assertGreater(node_db.updated_at, node_db.created_at)

        # Assert that the node from another host hasn't been touched
        # (updated_at is not updated)
        node_db = self._get_node_row(another_host_node)
        self.assertEqual(node_db.created_at, node_db.updated_at)

    def test_active_nodes(self):
        self._add_nodes_and_assert_exists(count=3)

        # Add another node from a different host
        with mock.patch.object(ovn_hash_ring_db, 'CONF') as mock_conf:
            mock_conf.host = 'another-host-' + uuidutils.generate_uuid()
            another_host_node = self._add_nodes_and_assert_exists()[0]

        # Assert all nodes are active (within 60 seconds)
        self.assertEqual(4, len(ovn_hash_ring_db.get_active_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP)))

        # Subtract 60 seconds from utcnow() and touch the nodes from our host
        time.sleep(1)
        fake_utcnow = timeutils.utcnow() - datetime.timedelta(seconds=60)
        with mock.patch.object(timeutils, 'utcnow') as mock_utcnow:
            mock_utcnow.return_value = fake_utcnow
            ovn_hash_ring_db.touch_nodes_from_host(self.admin_ctx,
                                                   HASH_RING_TEST_GROUP)

        # Now assert that all nodes from our host are seeing as offline.
        # Only the node from another host should be active
        active_nodes = ovn_hash_ring_db.get_active_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP)
        self.assertEqual(1, len(active_nodes))
        self.assertEqual(another_host_node, active_nodes[0].node_uuid)

    def test_active_nodes_from_host(self):
        self._add_nodes_and_assert_exists(count=3)

        # Add another node from a different host
        another_host_id = 'another-host-52359446-c366'
        with mock.patch.object(ovn_hash_ring_db, 'CONF') as mock_conf:
            mock_conf.host = another_host_id
            self._add_nodes_and_assert_exists()

        # Assert only the 3 nodes from this host is returned
        active_nodes = ovn_hash_ring_db.get_active_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP,
            from_host=True)
        self.assertEqual(3, len(active_nodes))
        self.assertNotIn(another_host_id, active_nodes)

    def test_touch_node(self):
        nodes = self._add_nodes_and_assert_exists(count=3)

        # Assert no nodes were updated yet
        for node in nodes:
            node_db = self._get_node_row(node)
            self.assertEqual(node_db.created_at, node_db.updated_at)

        # Touch one of the nodes
        time.sleep(1)
        ovn_hash_ring_db.touch_node(self.admin_ctx, nodes[0])

        # Assert it has been updated
        node_db = self._get_node_row(nodes[0])
        self.assertGreater(node_db.updated_at, node_db.created_at)

        # Assert the other two nodes hasn't been updated
        for node in nodes[1:]:
            node_db = self._get_node_row(node)
            self.assertEqual(node_db.created_at, node_db.updated_at)

    def test_active_nodes_different_groups(self):
        another_group = 'another_test_group'
        self._add_nodes_and_assert_exists(count=3)
        self._add_nodes_and_assert_exists(count=2, group_name=another_group)

        active_nodes = ovn_hash_ring_db.get_active_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP)
        self.assertEqual(3, len(active_nodes))
        for node in active_nodes:
            self.assertEqual(HASH_RING_TEST_GROUP, node.group_name)

        active_nodes = ovn_hash_ring_db.get_active_nodes(
            self.admin_ctx, interval=60, group_name=another_group)
        self.assertEqual(2, len(active_nodes))
        for node in active_nodes:
            self.assertEqual(another_group, node.group_name)

    def test_remove_nodes_from_host_different_groups(self):
        another_group = 'another_test_group'
        group1 = self._add_nodes_and_assert_exists(count=3)
        group2 = self._add_nodes_and_assert_exists(
            count=2, group_name=another_group)

        ovn_hash_ring_db.remove_nodes_from_host(self.admin_ctx,
                                                HASH_RING_TEST_GROUP)
        # Assert that all nodes from that group have been removed
        for node in group1:
            self.assertIsNone(self._get_node_row(node))

        # Assert that all nodes from a different group are intact
        for node in group2:
            self.assertIsNotNone(self._get_node_row(node))

    def test_touch_nodes_from_host_different_groups(self):
        another_group = 'another_test_group'
        group1 = self._add_nodes_and_assert_exists(count=3)
        group2 = self._add_nodes_and_assert_exists(
            count=2, group_name=another_group)

        # Assert that updated_at isn't updated yet
        for node in group1 + group2:
            node_db = self._get_node_row(node)
            self.assertEqual(node_db.created_at, node_db.updated_at)

        # Touch the nodes from group1
        time.sleep(1)
        ovn_hash_ring_db.touch_nodes_from_host(self.admin_ctx,
                                               HASH_RING_TEST_GROUP)

        # Assert that updated_at was updated for group1
        for node in group1:
            node_db = self._get_node_row(node)
            self.assertGreater(node_db.updated_at, node_db.created_at)

        # Assert that updated_at wasn't updated for group2
        for node in group2:
            node_db = self._get_node_row(node)
            self.assertEqual(node_db.created_at, node_db.updated_at)

    def test_count_offline_nodes(self):
        self._add_nodes_and_assert_exists(count=3)

        # Assert no nodes are considered offline
        self.assertEqual(0, ovn_hash_ring_db.count_offline_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP))

        # Subtract 60 seconds from utcnow() and touch the nodes to make
        # them to appear offline
        fake_utcnow = timeutils.utcnow() - datetime.timedelta(seconds=60)
        with mock.patch.object(timeutils, 'utcnow') as mock_utcnow:
            mock_utcnow.return_value = fake_utcnow
            ovn_hash_ring_db.touch_nodes_from_host(self.admin_ctx,
                                                   HASH_RING_TEST_GROUP)

        # Now assert that all nodes from our host are seeing as offline
        self.assertEqual(3, ovn_hash_ring_db.count_offline_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP))

        # Touch the nodes again without faking utcnow()
        ovn_hash_ring_db.touch_nodes_from_host(self.admin_ctx,
                                               HASH_RING_TEST_GROUP)

        # Assert no nodes are considered offline
        self.assertEqual(0, ovn_hash_ring_db.count_offline_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP))

    def test_remove_node_by_uuid(self):
        self._add_nodes_and_assert_exists(count=3)

        active_nodes = ovn_hash_ring_db.get_active_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP)
        self.assertEqual(3, len(active_nodes))

        node_to_remove = active_nodes[0].node_uuid
        ovn_hash_ring_db.remove_node_by_uuid(
                self.admin_ctx, node_to_remove)

        active_nodes = ovn_hash_ring_db.get_active_nodes(
            self.admin_ctx, interval=60, group_name=HASH_RING_TEST_GROUP)
        self.assertEqual(2, len(active_nodes))
        self.assertNotIn(node_to_remove, [n.node_uuid for n in active_nodes])

    def test_cleanup_old_nodes(self):
        # Add 2 new nodes
        self._add_nodes_and_assert_exists(count=2)

        # Subtract 5 days from utcnow() and touch the nodes to make
        # them to appear stale
        fake_utcnow = timeutils.utcnow() - datetime.timedelta(days=5)
        with mock.patch.object(timeutils, 'utcnow') as mock_utcnow:
            mock_utcnow.return_value = fake_utcnow
            ovn_hash_ring_db.touch_nodes_from_host(self.admin_ctx,
                                                   HASH_RING_TEST_GROUP)

        # Add 3 new nodes
        self._add_nodes_and_assert_exists(count=3)

        # Assert we have 5 nodes in the hash ring
        self.assertEqual(5, ovn_hash_ring_db.count_nodes_from_host(
            self.admin_ctx, HASH_RING_TEST_GROUP))

        # Clean up the 2 stale nodes
        ovn_hash_ring_db.cleanup_old_nodes(self.admin_ctx, days=5)

        # Assert we only have 3 node entries after the clean up
        self.assertEqual(3, ovn_hash_ring_db.count_nodes_from_host(
            self.admin_ctx, HASH_RING_TEST_GROUP))



# Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from neutron.tests.common.exclusive_resources import resource_allocator
from neutron.tests.common import net_helpers


class ExclusivePort(resource_allocator.ExclusiveResource):
    """Allocate a unique port for a specific protocol.

    :ivar port: allocated port
    :type port: int
    """

    def __init__(self, protocol, start=1024, end=None):
        super(ExclusivePort, self).__init__(
            'ports',
            functools.partial(net_helpers.get_free_namespace_port, protocol,
                              start=start, end=end))

    def _setUp(self):
        super(ExclusivePort, self)._setUp()
        self.port = self.resource



# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants

from neutron.agent.linux import ip_lib
from neutron.common import utils as common_utils
from neutron.plugins.ml2.drivers.macvtap.agent import macvtap_neutron_agent
from neutron.tests.common import net_helpers
from neutron.tests.functional import base as functional_base


class MacvtapAgentTestCase(functional_base.BaseSudoTestCase):
    def setUp(self):
        super(MacvtapAgentTestCase, self).setUp()
        self.mgr = macvtap_neutron_agent.MacvtapManager({})

    def test_get_all_devices(self):
        # NOTE(ralonsoh): Clean-up before testing. This test is executed with
        # concurrency=1. That means no other test is being executed at the same
        # time. Because the macvtap interface must be created in the root
        # namespace (``MacvtapManager`` cannot handle namespaces), the test
        # deletes any previous existing interface.
        for mac in self.mgr.get_all_devices():
            devices = ip_lib.IPWrapper().get_devices()
            for device in (d for d in devices if d.address == mac):
                device.link.delete()

        # Veth is simulating the hosts eth device. In this test it is used as
        # src_dev for the macvtap
        veth1, veth2 = self.useFixture(net_helpers.VethFixture()).ports
        macvtap = self.useFixture(net_helpers.MacvtapFixture(
            src_dev=veth1.name, mode='bridge',
            prefix=constants.MACVTAP_DEVICE_PREFIX)).ip_dev
        try:
            common_utils.wait_until_true(
                lambda: {macvtap.link.address} == self.mgr.get_all_devices(),
                timeout=10)
        except common_utils.WaitTimeout:
            msg = 'MacVTap address: %s, read devices: %s\n' % (
                macvtap.link.address, self.mgr.get_all_devices())
            for device in ip_lib.IPWrapper().get_devices():
                msg += '  Device %s, MAC: %s' % (device.name,
                                                 device.link.address)
            self.fail(msg)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for ml2 rpc
"""

import collections
from unittest import mock

from neutron_lib.agent import topics
from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.services.qos import constants as qos_consts
from oslo_config import cfg
from oslo_context import context as oslo_context
from sqlalchemy.orm import exc

from neutron.agent import rpc as agent_rpc
from neutron.db import provisioning_blocks
from neutron.plugins.ml2 import db as ml2_db
from neutron.plugins.ml2.drivers import type_tunnel
from neutron.plugins.ml2 import managers
from neutron.plugins.ml2 import rpc as plugin_rpc
from neutron.tests import base


cfg.CONF.import_group('ml2', 'neutron.conf.plugins.ml2.config')


class RpcCallbacksTestCase(base.BaseTestCase):

    def setUp(self):
        super(RpcCallbacksTestCase, self).setUp()
        self.type_manager = managers.TypeManager()
        self.notifier = plugin_rpc.AgentNotifierApi(topics.AGENT)
        self.callbacks = plugin_rpc.RpcCallbacks(self.notifier,
                                                 self.type_manager)
        self.plugin = mock.MagicMock()
        directory.add_plugin(plugin_constants.CORE, self.plugin)

    def _test_update_device_up(self, host=None):
        kwargs = {
            'agent_id': 'foo_agent',
            'device': 'foo_device',
            'host': host
        }
        with mock.patch('neutron.plugins.ml2.plugin.Ml2Plugin'
                        '._device_to_port_id'),\
                mock.patch.object(self.callbacks, 'notify_l2pop_port_wiring'):
            with mock.patch('neutron.db.provisioning_blocks.'
                            'provisioning_complete') as pc:
                self.callbacks.update_device_up(mock.Mock(), **kwargs)
                return pc

    def test_update_device_up_notify(self):
        notify = self._test_update_device_up()
        notify.assert_called_once_with(mock.ANY, mock.ANY, resources.PORT,
                                       provisioning_blocks.L2_AGENT_ENTITY)

    def test_update_device_up_notify_not_sent_with_port_not_found(self):
        self.plugin.port_bound_to_host.return_value = False
        notify = self._test_update_device_up('host')
        self.assertFalse(notify.call_count)

    def test_get_device_details_without_port_context(self):
        self.plugin.get_bound_port_context.return_value = None
        self.assertEqual(
            {'device': 'fake_device'},
            self.callbacks.get_device_details(mock.Mock(),
                                              device='fake_device'))

    def test_get_device_details_port_context_without_bounded_segment(self):
        self.plugin.get_bound_port_context().bottom_bound_segment = None
        self.assertEqual(
            {'device': 'fake_device'},
            self.callbacks.get_device_details(mock.Mock(),
                                              device='fake_device'))

    def test_get_device_details_port_status_equal_new_status(self):
        port = collections.defaultdict(lambda: 'fake')
        self.plugin.get_bound_port_context().current = port
        self.plugin.port_bound_to_host = port
        for admin_state_up in (True, False):
            new_status = (constants.PORT_STATUS_BUILD if admin_state_up
                          else constants.PORT_STATUS_DOWN)
            for status in (constants.PORT_STATUS_ACTIVE,
                           constants.PORT_STATUS_BUILD,
                           constants.PORT_STATUS_DOWN,
                           constants.PORT_STATUS_ERROR):
                port['admin_state_up'] = admin_state_up
                port['status'] = status
                self.plugin.update_port_status.reset_mock()
                self.callbacks.get_device_details(mock.Mock())
                self.assertEqual(status == new_status,
                                 not self.plugin.update_port_status.called)

    def test_get_device_details_caching(self):
        port = collections.defaultdict(lambda: 'fake_port')
        cached_networks = {}
        self.plugin.get_bound_port_context().current = port
        self.plugin.get_bound_port_context().network.current = (
            {"id": "fake_network"})
        self.callbacks.get_device_details(mock.Mock(), host='fake_host',
                                          cached_networks=cached_networks)
        self.assertIn('fake_port', cached_networks)

    def test_get_device_details_wrong_host(self):
        port = collections.defaultdict(lambda: 'fake')
        port_context = self.plugin.get_bound_port_context()
        port_context.current = port
        port_context.host = 'fake'
        self.plugin.update_port_status.reset_mock()
        self.callbacks.get_device_details(mock.Mock(),
                                          host='fake_host')
        self.assertFalse(self.plugin.update_port_status.called)

    def test_get_device_details_port_no_host(self):
        port = collections.defaultdict(lambda: 'fake')
        port_context = self.plugin.get_bound_port_context()
        port_context.current = port
        self.plugin.update_port_status.reset_mock()
        self.callbacks.get_device_details(mock.Mock())
        self.assertTrue(self.plugin.update_port_status.called)

    def test_get_device_details_qos_policy_id_none(self):
        port = collections.defaultdict(lambda: 'fake_port')
        self.plugin.get_bound_port_context().current = port
        self.plugin.get_bound_port_context().network._network = (
            {"id": "fake_network"})
        res = self.callbacks.get_device_details(mock.Mock(), host='fake')
        self.assertIsNone(res['qos_policy_id'])

    def test_get_device_details_network_qos_policy_id(self):
        port = collections.defaultdict(lambda: 'fake_port')
        self.plugin.get_bound_port_context().current = port
        self.plugin.get_bound_port_context().network._network = (
            {"id": "fake_network",
             qos_consts.QOS_POLICY_ID: 'test-policy-id'})
        res = self.callbacks.get_device_details(mock.Mock(), host='fake')
        self.assertEqual('test-policy-id', res['qos_network_policy_id'])

    def test_get_device_details_port_no_active_in_host(self):
        port = collections.defaultdict(lambda: 'fake_port')
        self.plugin.get_bound_port_context().current = port
        port['device_owner'] = constants.DEVICE_OWNER_COMPUTE_PREFIX
        port[portbindings.HOST_ID] = 'other-host'
        res = self.callbacks.get_device_details(mock.Mock(), host='host')
        self.assertIn(constants.NO_ACTIVE_BINDING, res)

    def test_get_device_details_qos_policy_id_from_port(self):
        port = collections.defaultdict(
            lambda: 'fake_port',
            {qos_consts.QOS_POLICY_ID: 'test-port-policy-id'})
        self.plugin.get_bound_port_context().current = port
        self.plugin.get_bound_port_context().network._network = (
            {"id": "fake_network",
             qos_consts.QOS_POLICY_ID: 'test-net-policy-id'})
        res = self.callbacks.get_device_details(mock.Mock(), host='fake')
        self.assertEqual('test-port-policy-id', res['qos_policy_id'])

    def _test_get_devices_list(self, callback, side_effect, expected):
        devices = [1, 2, 3, 4, 5]
        kwargs = {'host': 'fake_host', 'agent_id': 'fake_agent_id'}
        with mock.patch.object(self.callbacks, '_get_device_details',
                               side_effect=side_effect) as f:
            res = callback('fake_context', devices=devices, **kwargs)
            self.assertEqual(expected, res)
            self.assertEqual(len(devices), f.call_count)
            calls = [mock.call('fake_context', device=i,
                               port_context=mock.ANY, **kwargs)
                     for i in devices]
            f.assert_has_calls(calls)

    def test_get_devices_details_list(self):
        results = [{'device': [v]} for v in [1, 2, 3, 4, 5]]
        expected = results
        callback = self.callbacks.get_devices_details_list
        self._test_get_devices_list(callback, results, expected)

    def test_get_devices_details_list_with_empty_devices(self):
        with mock.patch.object(self.callbacks, 'get_device_details') as f:
            res = self.callbacks.get_devices_details_list('fake_context')
            self.assertFalse(f.called)
            self.assertEqual([], res)

    def test_get_devices_details_list_and_failed_devices(self):
        devices = [1, 2, 3, 4, 5]
        expected = {'devices': devices, 'failed_devices': []}
        callback = (
            self.callbacks.get_devices_details_list_and_failed_devices)
        self._test_get_devices_list(callback, devices, expected)

    def test_get_devices_details_list_and_failed_devices_failures(self):
        devices = [1, Exception('testdevice'), 3,
                   Exception('testdevice'), 5]
        expected = {'devices': [1, 3, 5], 'failed_devices': [2, 4]}
        callback = (
            self.callbacks.get_devices_details_list_and_failed_devices)
        self._test_get_devices_list(callback, devices, expected)

    def test_get_network_details(self):
        kwargs = {'agent_id': 'agent_id',
                  'host': 'host_id',
                  'network': 'network'}
        with mock.patch.object(self.plugin, 'get_network') as mock_get_network:
            mock_get_network.return_value = 'net_details'
            self.assertEqual(
                'net_details',
                self.callbacks.get_network_details('fake_context', **kwargs))
            mock_get_network.assert_called_once_with('fake_context', 'network')

    def test_get_devices_details_list_and_failed_devices_empty_dev(self):
        with mock.patch.object(self.callbacks, 'get_device_details') as f:
            res = self.callbacks.get_devices_details_list_and_failed_devices(
                'fake_context')
            self.assertFalse(f.called)
            self.assertEqual({'devices': [], 'failed_devices': []}, res)

    def _test_update_device_not_bound_to_host(self, func):
        self.plugin.port_bound_to_host.return_value = False
        self.callbacks.notify_l2pop_port_wiring = mock.Mock()
        self.plugin._device_to_port_id.return_value = 'fake_port_id'
        res = func(mock.Mock(), device='fake_device', host='fake_host')
        self.plugin.port_bound_to_host.assert_called_once_with(mock.ANY,
                                                               'fake_port_id',
                                                               'fake_host')
        return res

    def test_update_device_up_with_device_not_bound_to_host(self):
        with mock.patch.object(ml2_db, 'get_port') as ml2_db_get_port:
            self.assertIsNone(self._test_update_device_not_bound_to_host(
                self.callbacks.update_device_up))
            port = ml2_db_get_port.return_value
            (self.plugin.nova_notifier.notify_port_active_direct.
             assert_called_once_with(port))

    def test_update_device_up_with_device_not_bound_to_host_no_notify(self):
        cfg.CONF.set_override('notify_nova_on_port_status_changes', False)
        self.assertIsNone(self._test_update_device_not_bound_to_host(
            self.callbacks.update_device_up))
        self.plugin.nova_notifier.notify_port_active_direct.assert_not_called()

    def test_update_device_down_with_device_not_bound_to_host(self):
        self.assertEqual(
            {'device': 'fake_device', 'exists': True},
            self._test_update_device_not_bound_to_host(
                self.callbacks.update_device_down))

    def test_update_device_down_call_update_port_status(self):
        self.plugin.update_port_status.return_value = False
        self.callbacks.notify_l2pop_port_wiring = mock.Mock()
        self.plugin._device_to_port_id.return_value = 'fake_port_id'
        self.assertEqual(
            {'device': 'fake_device', 'exists': False},
            self.callbacks.update_device_down(mock.Mock(),
                                              device='fake_device',
                                              host='fake_host'))
        self.plugin.update_port_status.assert_called_once_with(
            mock.ANY, 'fake_port_id', constants.PORT_STATUS_DOWN,
            'fake_host')

    def test_update_device_down_call_update_port_status_failed(self):
        self.plugin.update_port_status.side_effect = exc.StaleDataError
        self.assertEqual({'device': 'fake_device', 'exists': False},
                         self.callbacks.update_device_down(
                             mock.Mock(), device='fake_device'))

    def _test_update_device_list(self, devices_up_side_effect,
                                 devices_down_side_effect, expected):
        devices_up = [1, 2, 3]
        devices_down = [4, 5]
        kwargs = {'host': 'fake_host', 'agent_id': 'fake_agent_id'}
        with mock.patch.object(self.callbacks, 'update_device_up',
                               side_effect=devices_up_side_effect) as f_up, \
            mock.patch.object(self.callbacks, 'update_device_down',
                              side_effect=devices_down_side_effect) as f_down:
            res = self.callbacks.update_device_list(
                'fake_context', devices_up=devices_up,
                devices_down=devices_down, **kwargs)
            self.assertEqual(expected, res)
            self.assertEqual(len(devices_up), f_up.call_count)
            self.assertEqual(len(devices_down), f_down.call_count)

    def test_update_device_list_no_failure(self):
        devices_up_side_effect = [1, 2, 3]
        devices_down_side_effect = [
             {'device': 4, 'exists': True},
             {'device': 5, 'exists': True}]
        expected = {'devices_up': devices_up_side_effect,
                    'failed_devices_up': [],
                    'devices_down':
                        [{'device': 4, 'exists': True},
                         {'device': 5, 'exists': True}],
                    'failed_devices_down': []}
        self._test_update_device_list(devices_up_side_effect,
                                      devices_down_side_effect,
                                      expected)

    def test_update_device_list_failed_devices(self):

        devices_up_side_effect = [1, Exception('testdevice'), 3]
        devices_down_side_effect = [{'device': 4, 'exists': True},
                        Exception('testdevice')]
        expected = {'devices_up': [1, 3],
                    'failed_devices_up': [2],
                    'devices_down':
                        [{'device': 4, 'exists': True}],
                    'failed_devices_down': [5]}

        self._test_update_device_list(devices_up_side_effect,
                                      devices_down_side_effect,
                                      expected)

    def test_update_device_list_empty_devices(self):

        expected = {'devices_up': [],
                    'failed_devices_up': [],
                    'devices_down': [],
                    'failed_devices_down': []}

        kwargs = {'host': 'fake_host', 'agent_id': 'fake_agent_id'}
        res = self.callbacks.update_device_list(
            'fake_context', devices_up=[], devices_down=[], **kwargs)
        self.assertEqual(expected, res)


class RpcApiTestCase(base.BaseTestCase):

    def _test_rpc_api(self, rpcapi, topic, method, rpc_method, **kwargs):
        if method == "update_device_list":
            expected = {'devices_up': [],
                        'failed_devices_up': [],
                        'devices_down': [],
                        'failed_devices_down': []}
        else:
            expected = 'foo'

        ctxt = oslo_context.RequestContext(user_id='fake_user',
                                           project_id='fake_project')
        expected_retval = expected if rpc_method == 'call' else None
        expected_version = kwargs.pop('version', None)
        fanout = kwargs.pop('fanout', False)

        with mock.patch.object(rpcapi.client, rpc_method) as rpc_mock,\
                mock.patch.object(rpcapi.client, 'prepare') as prepare_mock:
            prepare_mock.return_value = rpcapi.client
            rpc_mock.return_value = expected_retval
            retval = getattr(rpcapi, method)(ctxt, **kwargs)

        prepare_args = {}
        if expected_version:
            prepare_args['version'] = expected_version
        if fanout:
            prepare_args['fanout'] = fanout
        if topic:
            prepare_args['topic'] = topic
        prepare_mock.assert_called_once_with(**prepare_args)

        self.assertEqual(retval, expected_retval)
        rpc_mock.assert_called_once_with(ctxt, method, **kwargs)

    def test_delete_network(self):
        rpcapi = plugin_rpc.AgentNotifierApi(topics.AGENT)
        self._test_rpc_api(
                rpcapi,
                topics.get_topic_name(topics.AGENT,
                                      topics.NETWORK,
                                      topics.DELETE),
                'network_delete', rpc_method='cast',
                fanout=True, network_id='fake_request_spec')

    def test_port_update(self):
        rpcapi = plugin_rpc.AgentNotifierApi(topics.AGENT)
        self._test_rpc_api(
                rpcapi,
                topics.get_topic_name(topics.AGENT,
                                      topics.PORT,
                                      topics.UPDATE),
                'port_update', rpc_method='cast',
                fanout=True, port='fake_port',
                network_type='fake_network_type',
                segmentation_id='fake_segmentation_id',
                physical_network='fake_physical_network')

    def test_port_delete(self):
        rpcapi = plugin_rpc.AgentNotifierApi(topics.AGENT)
        self._test_rpc_api(
            rpcapi,
            topics.get_topic_name(topics.AGENT,
                                  topics.PORT,
                                  topics.DELETE),
            'port_delete', rpc_method='cast',
            fanout=True, port_id='fake_port')

    def test_tunnel_update(self):
        rpcapi = plugin_rpc.AgentNotifierApi(topics.AGENT)
        self._test_rpc_api(
                rpcapi,
                topics.get_topic_name(topics.AGENT,
                                      type_tunnel.TUNNEL,
                                      topics.UPDATE),
                'tunnel_update', rpc_method='cast',
                fanout=True,
                tunnel_ip='fake_ip', tunnel_type='gre')

    def test_tunnel_delete(self):
        rpcapi = plugin_rpc.AgentNotifierApi(topics.AGENT)
        self._test_rpc_api(
                rpcapi,
                topics.get_topic_name(topics.AGENT,
                                      type_tunnel.TUNNEL,
                                      topics.DELETE),
                'tunnel_delete', rpc_method='cast',
                fanout=True,
                tunnel_ip='fake_ip', tunnel_type='gre')

    def test_device_details(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'get_device_details', rpc_method='call',
                           device='fake_device',
                           agent_id='fake_agent_id',
                           host='fake_host',
                           version='1.9')

    def test_devices_details_list(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'get_devices_details_list', rpc_method='call',
                           devices=['fake_device1', 'fake_device2'],
                           agent_id='fake_agent_id', host='fake_host',
                           version='1.9')

    def test_update_device_down(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'update_device_down', rpc_method='call',
                           device='fake_device',
                           agent_id='fake_agent_id',
                           host='fake_host',
                           version='1.9')

    def test_tunnel_sync(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'tunnel_sync', rpc_method='call',
                           tunnel_ip='fake_tunnel_ip',
                           tunnel_type=None,
                           host='fake_host',
                           version='1.4')

    def test_update_device_up(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'update_device_up', rpc_method='call',
                           device='fake_device',
                           agent_id='fake_agent_id',
                           host='fake_host',
                           version='1.9')

    def test_update_device_list(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'update_device_list', rpc_method='call',
                           devices_up=['fake_device1', 'fake_device2'],
                           devices_down=['fake_device3', 'fake_device4'],
                           agent_id='fake_agent_id',
                           host='fake_host',
                           refresh_tunnels=False,
                           version='1.9')

    def test_get_devices_details_list_and_failed_devices(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'get_devices_details_list_and_failed_devices',
                           rpc_method='call',
                           devices=['fake_device1', 'fake_device2'],
                           agent_id='fake_agent_id',
                           host='fake_host',
                           version='1.5')

    def test_devices_details_list_and_failed_devices(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'get_devices_details_list_and_failed_devices',
                           rpc_method='call',
                           devices=['fake_device1', 'fake_device2'],
                           agent_id='fake_agent_id', host='fake_host',
                           version='1.5')

    def test_get_ports_by_vnic_type_and_host(self):
        rpcapi = agent_rpc.PluginApi(topics.PLUGIN)
        self._test_rpc_api(rpcapi, None,
                           'get_ports_by_vnic_type_and_host',
                           rpc_method='call',
                           vnic_type='fake_device1',
                           host='fake_host',
                           version='1.7')



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from neutron_lib import policy as neutron_policy
from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base

DEPRECATED_REASON = (
    "The security group API now supports system scope and default roles.")


SG_COLLECTION_PATH = '/security-groups'
SG_RESOURCE_PATH = '/security-groups/{id}'
SG_TAGS_PATH = SG_RESOURCE_PATH + '/tags'
SG_TAG_PATH = SG_RESOURCE_PATH + '/tags/{tag_id}'
RULE_COLLECTION_PATH = '/security-group-rules'
RULE_RESOURCE_PATH = '/security-group-rules/{id}'
RULE_TAGS_PATH = RULE_RESOURCE_PATH + '/tags'
RULE_TAG_PATH = RULE_RESOURCE_PATH + '/tags/{tag_id}'

RULE_ADMIN_OR_SG_OWNER = 'rule:admin_or_sg_owner'
RULE_ADMIN_OWNER_OR_SG_OWNER = 'rule:admin_owner_or_sg_owner'


SG_ACTION_GET_TAGS = [
    {'method': 'GET', 'path': SG_TAGS_PATH},
    {'method': 'GET', 'path': SG_TAG_PATH},
]
SG_ACTION_PUT_TAGS = [
    {'method': 'PUT', 'path': SG_TAGS_PATH},
    {'method': 'PUT', 'path': SG_TAG_PATH},
]
SG_ACTION_DELETE_TAGS = [
    {'method': 'DELETE', 'path': SG_TAGS_PATH},
    {'method': 'DELETE', 'path': SG_TAG_PATH},
]


rules = [
    policy.RuleDefault(
        name='admin_or_sg_owner',
        check_str=neutron_policy.policy_or(
            'rule:context_is_admin',
            'tenant_id:%(security_group:tenant_id)s'),
        description='Rule for admin or security group owner access'),
    policy.RuleDefault(
        name='admin_owner_or_sg_owner',
        check_str=neutron_policy.policy_or(
            'rule:owner',
            RULE_ADMIN_OR_SG_OWNER),
        description=('Rule for resource owner, '
                     'admin or security group owner access')),
    policy.RuleDefault(
        name='shared_security_group',
        check_str='field:security_groups:shared=True',
        description='Definition of a shared security group'
    ),
    policy.RuleDefault(
        name='rule_default_sg',
        check_str='field:security_group_rules:belongs_to_default_sg=True',
        description='Definition of a security group rule that belongs to the '
                    'project default security group'
    ),

    # TODO(amotoki): admin_or_owner is the right rule?
    # Does an empty string make more sense for create_security_group?
    policy.DocumentedRuleDefault(
        name='create_security_group',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Create a security group',
        operations=[
            {
                'method': 'POST',
                'path': SG_COLLECTION_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_security_group',
            check_str=neutron_policy.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_security_group',
        check_str=neutron_policy.policy_or(
            base.ADMIN_OR_PROJECT_READER,
            'rule:shared_security_group'
        ),
        scope_types=['project'],
        description='Get a security group',
        operations=[
            {
                'method': 'GET',
                'path': SG_COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': SG_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_security_group',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_security_groups_tags',
        check_str=neutron_policy.policy_or(
            base.ADMIN_OR_PROJECT_READER,
            'rule:shared_security_group'
        ),
        scope_types=['project'],
        description='Get the security group tags',
        operations=SG_ACTION_GET_TAGS,
    ),
    policy.DocumentedRuleDefault(
        name='update_security_group',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Update a security group',
        operations=[
            {
                'method': 'PUT',
                'path': SG_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_security_group',
            check_str=neutron_policy.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_security_groups_tags',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Update the security group tags',
        operations=SG_ACTION_PUT_TAGS,
    ),
    policy.DocumentedRuleDefault(
        name='delete_security_group',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Delete a security group',
        operations=[
            {
                'method': 'DELETE',
                'path': SG_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_security_group',
            check_str=neutron_policy.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_security_groups_tags',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Delete the security group tags',
        operations=SG_ACTION_DELETE_TAGS,
    ),

    # TODO(amotoki): admin_or_owner is the right rule?
    # Does an empty string make more sense for create_security_group_rule?
    policy.DocumentedRuleDefault(
        name='create_security_group_rule',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Create a security group rule',
        operations=[
            {
                'method': 'POST',
                'path': RULE_COLLECTION_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_security_group_rule',
            check_str=neutron_policy.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_security_group_rule',
        check_str=neutron_policy.policy_or(
            base.ADMIN_OR_PROJECT_READER,
            base.RULE_SG_OWNER),
        scope_types=['project'],
        description='Get a security group rule',
        operations=[
            {
                'method': 'GET',
                'path': RULE_COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': RULE_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_security_group_rule',
            check_str=RULE_ADMIN_OWNER_OR_SG_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_security_group_rule',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description='Delete a security group rule',
        operations=[
            {
                'method': 'DELETE',
                'path': RULE_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_security_group_rule',
            check_str=neutron_policy.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
]


def list_rules():
    return rules



# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes
from ctypes import util as ctypes_util

from pyroute2 import netlink


_CDLL = None


def get_cdll():
    global _CDLL
    if not _CDLL:
        # NOTE(ralonsoh): from https://docs.python.org/3.6/library/
        # ctypes.html#ctypes.PyDLL: "Instances of this class behave like CDLL
        # instances, except that the Python GIL is not released during the
        # function call, and after the function execution the Python error
        # flag is checked."
        # Check https://bugs.launchpad.net/neutron/+bug/1870352
        _CDLL = ctypes.PyDLL(ctypes_util.find_library('c'), use_errno=True)
    return _CDLL


def make_serializable(value):
    """Make a pyroute2 object serializable

    This function converts 'netlink.nla_slot' object (key, value) in a list
    of two elements.
    """
    def _ensure_string(value):
        return value.decode() if isinstance(value, bytes) else value

    if isinstance(value, list):
        return [make_serializable(item) for item in value]
    elif isinstance(value, netlink.nla_slot):
        return [_ensure_string(value[0]), make_serializable(value[1])]
    elif isinstance(value, netlink.nla_base):
        return make_serializable(value.dump())
    elif isinstance(value, dict):
        return {_ensure_string(key): make_serializable(data)
                for key, data in value.items()}
    elif isinstance(value, tuple):
        return tuple(make_serializable(item) for item in value)
    return _ensure_string(value)



# (c) Copyright 2014 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_log import log as logging

from neutron.api import extensions
from neutron.api.v2 import base
from neutron.quota import resource_registry

LOG = logging.getLogger(__name__)


def build_plural_mappings(special_mappings, resource_map):
    """Create plural to singular mapping for all resources.

    Allows for special mappings to be provided, for particular cases..
    Otherwise, will strip off the last character for normal mappings, like
    routers -> router, unless the plural name ends with 'ies', in which
    case the singular form will end with a 'y' (e.g.: policy/policies)
    """
    plural_mappings = {}
    for plural in resource_map:
        singular = special_mappings.get(plural)
        if not singular:
            if plural.endswith('ies'):
                singular = "%sy" % plural[:-3]
            else:
                singular = plural[:-1]
        plural_mappings[plural] = singular
    return plural_mappings


def build_resource_info(plural_mappings, resource_map, which_service,
                        action_map=None, register_quota=False,
                        translate_name=False, allow_bulk=False):
    """Build resources for advanced services.

    Takes the resource information, and singular/plural mappings, and creates
    API resource objects for advanced services extensions. Will optionally
    translate underscores to dashes in resource names, register the resource,
    and accept action information for resources.

    :param plural_mappings: mappings between singular and plural forms
    :param resource_map: attribute map for the WSGI resources to create
    :param which_service: The name of the service for which the WSGI resources
                          are being created. This name will be used to pass
                          the appropriate plugin to the WSGI resource.
                          It can be set to None or "CORE" to create WSGI
                          resources for the core plugin
    :param action_map: custom resource actions
    :param register_quota: it can be set to True to register quotas for the
                           resource(s) being created
    :param translate_name: replaces underscores with dashes
    :param allow_bulk: True if bulk create are allowed
    """
    resources = []
    if not which_service:
        which_service = constants.CORE
    if action_map is None:
        action_map = {}
    plugin = directory.get_plugin(which_service)
    path_prefix = getattr(plugin, "path_prefix", "")
    LOG.debug('Service %(service)s assigned prefix: %(prefix)s',
              {'service': which_service, 'prefix': path_prefix})
    for collection_name in resource_map:
        resource_name = plural_mappings[collection_name]
        params = resource_map.get(collection_name, {})
        # If SUB_RESOURCE_ATTRIBUTE_MAP was passed in as a resource_map, we
        # need special handling for it. SUB_RESOURCE_ATTRIBUTE_MAP must have
        # a 'parent' and 'parameters' keys. 'parameters' key is going to
        # contain sub-resources that are being extended.
        parent = params.get('parent')
        params = params['parameters'] if params.get('parameters') else params

        if translate_name:
            collection_name = collection_name.replace('_', '-')
        if register_quota:
            resource_registry.register_resource_by_name(resource_name)
        member_actions = action_map.get(resource_name, {})
        controller = base.create_resource(
            collection_name, resource_name, plugin, params,
            member_actions=member_actions,
            allow_bulk=allow_bulk,
            allow_pagination=True,
            allow_sorting=True,
            parent=parent)
        resource = extensions.ResourceExtension(
            collection_name,
            controller,
            path_prefix=path_prefix,
            member_actions=member_actions,
            attr_map=params,
            parent=parent)
        resources.append(resource)
    return resources



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import context
from oslo_utils import uuidutils

from neutron.db import rbac_db_models
from neutron.objects import base as obj_base
from neutron.objects.db import api as obj_db_api
from neutron.objects import network as net_obj
from neutron.objects import rbac_db
from neutron.objects import subnet
from neutron.tests.unit.objects import test_base as obj_test_base
from neutron.tests.unit import testlib_api


class IPAllocationPoolObjectIfaceTestCase(
        obj_test_base.BaseObjectIfaceTestCase):

    _test_class = subnet.IPAllocationPool


class IPAllocationPoolDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                                       testlib_api.SqlTestCase):

    _test_class = subnet.IPAllocationPool

    def setUp(self):
        super(IPAllocationPoolDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'subnet_id': lambda: self._create_test_subnet_id()})


class DNSNameServerObjectIfaceTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = subnet.DNSNameServer

    def setUp(self):
        super(DNSNameServerObjectIfaceTestCase, self).setUp()
        self.pager_map[self._test_class.obj_name()] = (
            obj_base.Pager(sorts=[('order', True)]))


class DNSNameServerDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                                    testlib_api.SqlTestCase):

    _test_class = subnet.DNSNameServer

    def setUp(self):
        super(DNSNameServerDbObjectTestCase, self).setUp()
        self._subnet_id = self._create_test_subnet_id()
        self.update_obj_fields({'subnet_id': self._subnet_id})

    def _create_dnsnameservers(self):
        for obj in self.obj_fields:
            dns = self._make_object(obj)
            dns.create()

    def test_get_objects_sort_by_order_asc(self):
        self._create_dnsnameservers()
        objs = self._test_class.get_objects(self.context)
        fields_sorted = sorted([obj['order'] for obj in self.obj_fields])
        self.assertEqual(fields_sorted, [obj.order for obj in objs])

    def test_get_objects_sort_by_order_desc(self):
        self._create_dnsnameservers()
        pager = obj_base.Pager(sorts=[('order', False)])
        objs = self._test_class.get_objects(self.context, _pager=pager,
                                            subnet_id=self._subnet_id)
        fields_sorted = sorted([obj['order'] for obj in self.obj_fields],
                               reverse=True)
        self.assertEqual(fields_sorted, [obj.order for obj in objs])

    def test_get_objects_sort_by_address_asc_using_pager(self):
        self._create_dnsnameservers()
        pager = obj_base.Pager(sorts=[('address', True)])
        objs = self._test_class.get_objects(self.context, _pager=pager)
        fields_sorted = sorted([obj['address'] for obj in self.obj_fields])
        self.assertEqual(fields_sorted, [obj.address for obj in objs])


class RouteObjectIfaceTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = subnet.Route


class RouteDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                            testlib_api.SqlTestCase):

    _test_class = subnet.Route

    def setUp(self):
        super(RouteDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'subnet_id': lambda: self._create_test_subnet_id()})


class SubnetServiceTypeObjectIfaceTestCase(
        obj_test_base.BaseObjectIfaceTestCase):

    _test_class = subnet.SubnetServiceType


class SubnetServiceTypeDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                            testlib_api.SqlTestCase):

    _test_class = subnet.SubnetServiceType

    def setUp(self):
        super(SubnetServiceTypeDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'subnet_id': lambda: self._create_test_subnet_id()})


class SubnetObjectIfaceTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = subnet.Subnet

    def setUp(self):
        super(SubnetObjectIfaceTestCase, self).setUp()
        self.pager_map[subnet.DNSNameServer.obj_name()] = (
            obj_base.Pager(sorts=[('order', True)]))
        # Base class will mock those out only when rbac_db_model is set for the
        # object. Since subnets don't have their own models but only derive
        # shared value from networks, we need to unconditionally mock those
        # entry points out here, otherwise they will trigger database access,
        # which is not allowed in 'Iface' test classes.
        mock.patch.object(
            rbac_db.RbacNeutronDbObjectMixin,
            'is_shared_with_project', return_value=False).start()
        mock.patch.object(
            rbac_db.RbacNeutronDbObjectMixin,
            'get_shared_with_project').start()


class SubnetDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                             testlib_api.SqlTestCase):

    _test_class = subnet.Subnet

    CORE_PLUGIN = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'

    def setUp(self):
        super(SubnetDbObjectTestCase, self).setUp()
        # set up plugin because some models used here require a plugin
        # (specifically, rbac models and their get_valid_actions validators)
        self.setup_coreplugin(self.CORE_PLUGIN)
        network_id = self._create_test_network_id()
        self.update_obj_fields(
            {'network_id': network_id,
             'segment_id': lambda: self._create_test_segment_id(network_id)})

    def test_get_dns_nameservers_in_order(self):
        obj = self._make_object(self.obj_fields[0])
        obj.create()
        dns_nameservers = [(2, '1.2.3.4'), (1, '5.6.7.8'), (4, '7.7.7.7')]
        for order, address in dns_nameservers:
            dns = subnet.DNSNameServer(self.context, order=order,
                                       address=address,
                                       subnet_id=obj.id)
            dns.create()

        new = self._test_class.get_object(self.context, id=obj.id)
        self.assertEqual(1, new.dns_nameservers[0].order)
        self.assertEqual(2, new.dns_nameservers[1].order)
        self.assertEqual(4, new.dns_nameservers[-1].order)

    def _create_shared_network_rbac_entry(self, network):
        attrs = {
            'object_id': network['id'],
            'target_tenant': '*',
            'action': rbac_db_models.ACCESS_SHARED
        }
        obj_db_api.create_object(net_obj.NetworkRBAC, self.context, attrs)

    def test_get_subnet_shared_true(self):
        network = self._create_test_network()
        self._create_shared_network_rbac_entry(network)
        subnet_data = dict(self.obj_fields[0])
        subnet_data['network_id'] = network['id']

        obj = self._make_object(subnet_data)
        # check if shared will be load by 'obj_load_attr' and using extra query
        # by RbacNeutronDbObjectMixin get_shared_with_project
        self.assertTrue(obj.shared)
        obj.create()
        # here the shared should be load by is_network_shared
        self.assertTrue(obj.shared)

        new = self._test_class.get_object(self.context,
                                          **obj._get_composite_keys())
        # again, the shared should be load by is_network_shared
        self.assertTrue(new.shared)

    def test_filter_by_shared(self):
        network = self._create_test_network()
        self._create_shared_network_rbac_entry(network)

        subnet_data = dict(self.obj_fields[0])
        subnet_data['network_id'] = network['id']
        obj = self._make_object(subnet_data)
        obj.create()

        result = self._test_class.get_objects(self.context, shared=True)

        self.assertEqual(obj, result[0])

    def test_get_shared_subnet_with_another_tenant(self):
        network_shared = self._create_test_network()
        self._create_shared_network_rbac_entry(network_shared)

        subnet_data = dict(self.obj_fields[0])
        subnet_data['network_id'] = network_shared['id']
        shared_subnet = self._make_object(subnet_data)
        shared_subnet.create()

        priv_subnet = self._make_object(self.obj_fields[1])
        priv_subnet.create()

        # Situation here:
        #   - we have one network with a subnet that are private
        #   - shared network with its subnet
        # creating new context, user should have access to one shared network

        all_subnets = self._test_class.get_objects(self.context)
        self.assertEqual(2, len(all_subnets))

        # access with new tenant_id, should be able to access to one subnet
        new_ctx = context.Context('', uuidutils.generate_uuid())
        public_subnets = self._test_class.get_objects(new_ctx)
        self.assertEqual([shared_subnet], public_subnets)

        # test get_object to fetch the private and then the shared subnet
        fetched_private_subnet = self._test_class.get_object(new_ctx,
                                                             id=priv_subnet.id)
        self.assertIsNone(fetched_private_subnet)

        fetched_public_subnet = (
            self._test_class.get_object(new_ctx, id=shared_subnet.id))
        self.assertEqual(shared_subnet, fetched_public_subnet)

    def test_get_service_types(self):
        obj = self._make_object(self.obj_fields[0])
        obj.create()

        service_type_obj = subnet.SubnetServiceType(
            self.context, subnet_id=obj.id, service_type='dhcp-agent')
        service_type_obj.create()

        listed_obj = subnet.Subnet.get_object(self.context, id=obj.id)

        self.assertEqual([service_type_obj.service_type],
                         listed_obj.service_types)

        # Try to load the service_types by obj_load_attr
        obj1 = self._make_object(self.obj_fields[0])
        self.assertEqual([service_type_obj.service_type],
                         obj1.service_types)

    def test_find_candidate_subnets(self):
        network = self._create_test_network()
        subnet_data = dict(self.obj_fields[0])
        subnet_data['network_id'] = network['id']
        subnet_net = self._make_object(subnet_data)
        subnet2_data = dict(self.obj_fields[0])
        subnet2_data['id'] = uuidutils.generate_uuid()
        subnet2_data['network_id'] = network['id']
        subnet2_net = self._make_object(subnet2_data)
        subnet_net.create()
        subnet2_net.create()
        fixed_ips = [
            {'subnet_id': subnet_data['id'], 'ip_address': '10.0.0.2'},
            {'subnet_id': subnet2_data['id'], 'ip_address': '10.0.1.2'}]
        candidate_subnet = subnet.Subnet.find_candidate_subnets(
            self.context, network['id'], None, None, True, fixed_ips)
        self.assertEqual(2, len(candidate_subnet))
        self.assertNotEqual(
            candidate_subnet[0]['id'], candidate_subnet[1]['id'])


class NetworkSubnetLockTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = subnet.NetworkSubnetLock


class NetworkSubnetLockDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                                        testlib_api.SqlTestCase):

    _test_class = subnet.NetworkSubnetLock

    def setUp(self):
        super(NetworkSubnetLockDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'network_id': lambda: self._create_test_network_id()})

    def test_lock_subnet_update(self):
        obj = self._make_object(self.obj_fields[0])
        obj.create()
        subnet_id = self._create_test_subnet_id(network_id=obj.network_id)
        subnet.NetworkSubnetLock.lock_subnet(self.context, obj.network_id,
                                             subnet_id)
        obj = subnet.NetworkSubnetLock.get_object(self.context,
                                                  network_id=obj.network_id)
        self.assertEqual(subnet_id, obj.subnet_id)

    def test_lock_subnet_create(self):
        network_id = self._create_test_network_id()
        subnet_id = self._create_test_subnet_id(network_id=network_id)
        obj = subnet.NetworkSubnetLock.get_object(self.context,
                                                  network_id=network_id)
        self.assertIsNone(obj)

        subnet.NetworkSubnetLock.lock_subnet(self.context, network_id,
                                             subnet_id)
        obj = subnet.NetworkSubnetLock.get_object(self.context,
                                                  network_id=network_id)
        self.assertEqual(network_id, obj.network_id)
        self.assertEqual(subnet_id, obj.subnet_id)



# Copyright 2011 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools
import itertools
import random
import re
import time
import uuid

from neutron_lib import constants as p_const
from neutron_lib import exceptions
from neutron_lib.plugins.ml2 import ovs_constants
from neutron_lib.services.qos import constants as qos_constants
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import exceptions as ovs_exceptions

import tenacity

from neutron._i18n import _
from neutron.agent.common import ip_lib
from neutron.agent.common import utils
from neutron.agent.ovsdb import impl_idl
from neutron.common import _constants as common_constants
from neutron.common import utils as common_utils
from neutron.conf.agent import ovs_conf
from neutron.plugins.ml2.drivers.openvswitch.agent.common \
    import exceptions as ovs_exc


UINT64_BITMASK = (1 << 64) - 1

# Special return value for an invalid OVS ofport
INVALID_OFPORT = -1
UNASSIGNED_OFPORT = []

# OVS bridge fail modes
FAILMODE_SECURE = 'secure'
FAILMODE_STANDALONE = 'standalone'

# special values for cookies
COOKIE_ANY = object()

ovs_conf.register_ovs_agent_opts()

LOG = logging.getLogger(__name__)

OVS_DEFAULT_CAPS = {
    'datapath_types': [],
    'iface_types': [],
}

# It's default queue, all packets not tagged with 'set_queue' will go through
# this one
QOS_DEFAULT_QUEUE = 0

_SENTINEL = object()

CTRL_RATE_LIMIT_MIN = 100
CTRL_BURST_LIMIT_MIN = 25
OVS_MAX_RATE = 2 ** 35 - 1

# TODO(slaweq): move this to neutron_lib.constants
TYPE_GRE_IP6 = 'ip6gre'

ActionFlowTuple = collections.namedtuple('ActionFlowTuple',
                                         ['action', 'flow', 'flow_group_id'])


MAX_METER_REGEX = re.compile(r"max_meter:(\w+) max_bands:(\w+) *")
BAND_TYPES_REGEX = re.compile(r"band_types: (\w+)")
CAPS_REGEX = re.compile(r"capabilities: (\w+) (\w+) (\w+) (\w+)")


def _ovsdb_result_pending(result):
    """Return True if ovsdb indicates the result is still pending."""
    # ovsdb can return '[]' for an ofport that has not yet been assigned
    return result == []


def _ovsdb_retry(fn):
    """Decorator for retrying when OVS has yet to assign an ofport.

    The instance's ovsdb_timeout is used as the max waiting time. This relies
    on the fact that instance methods receive self as the first argument.
    """
    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        self = args[0]
        new_fn = tenacity.retry(
            reraise=True,
            retry=tenacity.retry_if_result(_ovsdb_result_pending),
            wait=tenacity.wait_exponential(multiplier=0.02, max=1),
            stop=tenacity.stop_after_delay(
                self.ovsdb_timeout))(fn)
        return new_fn(*args, **kwargs)
    return wrapped


def get_gre_tunnel_port_type(remote_ip, local_ip):
    if (common_utils.get_ip_version(remote_ip) == p_const.IP_VERSION_6 or
            common_utils.get_ip_version(local_ip) == p_const.IP_VERSION_6):
        return TYPE_GRE_IP6
    return p_const.TYPE_GRE


class VifPort(object):
    def __init__(self, port_name, ofport, vif_id, vif_mac, switch):
        self.port_name = port_name
        self.ofport = ofport
        self.vif_id = vif_id
        self.vif_mac = vif_mac
        self.switch = switch

    def __str__(self):
        return ("iface-id=%s, vif_mac=%s, port_name=%s, ofport=%s, "
                "bridge_name=%s") % (
                    self.vif_id, self.vif_mac,
                    self.port_name, self.ofport,
                    self.switch.br_name)


class BaseOVS(object):

    def __init__(self):
        self.ovsdb = impl_idl.api_factory()
        self._hw_offload = None

    @property
    def ovsdb_timeout(self):
        return self.ovsdb.ovsdb_connection.timeout

    def _execute(self, cmd, check_error=False, log_errors=True):
        try:
            return cmd.execute(check_error=True, log_errors=False)
        except Exception as e:
            if log_errors:
                (LOG.warning
                 if (isinstance(e, ovs_exceptions.OvsdbAppException) and
                     not check_error)
                 else LOG.exception)("Error while querying OVSDB: %s", e)
            if check_error:
                raise

    def add_bridge(self, bridge_name,
                   datapath_type=ovs_constants.OVS_DATAPATH_SYSTEM):
        br = OVSBridge(bridge_name, datapath_type=datapath_type)
        br.create()
        return br

    def delete_bridge(self, bridge_name):
        self.ovsdb.del_br(bridge_name).execute()

    def bridge_exists(self, bridge_name):
        return self.ovsdb.br_exists(bridge_name).execute()

    def port_exists(self, port_name):
        cmd = self.ovsdb.db_get('Port', port_name, 'name')
        return bool(cmd.execute(check_error=False, log_errors=False))

    def get_bridge_for_iface(self, iface):
        return self.ovsdb.iface_to_br(iface).execute()

    def get_bridges(self):
        return self.ovsdb.list_br().execute(check_error=True)

    def get_bridge_external_bridge_id(self, bridge, check_error=False,
                                      log_errors=True):
        return self._execute(
            self.ovsdb.br_get_external_id(bridge, 'bridge-id'),
            check_error=check_error, log_errors=log_errors)

    def set_db_attribute(self, table_name, record, column, value,
                         check_error=False, log_errors=True):
        self._execute(
            self.ovsdb.db_set(table_name, record, (column, value)),
            check_error=check_error, log_errors=log_errors)

    def clear_db_attribute(self, table_name, record, column):
        self.ovsdb.db_clear(table_name, record, column).execute()

    def db_get_val(self, table, record, column, check_error=False,
                   log_errors=True):
        return self._execute(
            self.ovsdb.db_get(table, record, column),
            check_error=check_error, log_errors=log_errors)

    @property
    def config(self):
        """A dict containing the only row from the root Open_vSwitch table

        This row contains several columns describing the Open vSwitch install
        and the system on which it is installed. Useful keys include:
            datapath_types: a list of supported datapath types
            iface_types: a list of supported interface types
            ovs_version: the OVS version
        """
        return self.ovsdb.db_list("Open_vSwitch").execute()[0]

    @property
    def capabilities(self):
        _cfg = self.config
        return {k: _cfg.get(k, OVS_DEFAULT_CAPS[k]) for k in OVS_DEFAULT_CAPS}

    @property
    def is_hw_offload_enabled(self):
        if self._hw_offload is None:
            self._hw_offload = self.config.get('other_config', {}).get(
                'hw-offload', '').lower() == 'true'
        return self._hw_offload


# Map from version string to on-the-wire protocol version encoding:
OF_PROTOCOL_TO_VERSION = {
    ovs_constants.OPENFLOW10: 1,
    ovs_constants.OPENFLOW11: 2,
    ovs_constants.OPENFLOW12: 3,
    ovs_constants.OPENFLOW13: 4,
    ovs_constants.OPENFLOW14: 5,
    ovs_constants.OPENFLOW15: 6,
}


def version_from_protocol(protocol):
    if protocol not in OF_PROTOCOL_TO_VERSION:
        raise Exception(_("unknown OVS protocol string, cannot compare: "
                          "%(protocol)s, (known: %(known)s)") %
                        {'protocol': protocol,
                         'known': list(OF_PROTOCOL_TO_VERSION)})
    return OF_PROTOCOL_TO_VERSION[protocol]


class OVSBridge(BaseOVS):
    def __init__(self, br_name,
                 datapath_type=ovs_constants.OVS_DATAPATH_SYSTEM):
        super(OVSBridge, self).__init__()
        self.br_name = br_name
        self.datapath_type = datapath_type
        self._default_cookie = generate_random_cookie()
        self._highest_protocol_needed = ovs_constants.OPENFLOW10
        self._min_bw_qos_id = uuidutils.generate_uuid()
        # TODO(jlibosva): Revert initial_protocols once launchpad bug 1852221
        #                 is fixed and new openvswitch containing the fix is
        #                 released.
        self.initial_protocols = {
            ovs_constants.OPENFLOW10,
            ovs_constants.OPENFLOW13,
            ovs_constants.OPENFLOW14}
        self.initial_protocols.add(self._highest_protocol_needed)
        self._flows_per_port = cfg.CONF.OVS.openflow_processed_per_port

    @property
    def default_cookie(self):
        return self._default_cookie

    def set_agent_uuid_stamp(self, val):
        self._default_cookie = val

    def disable_in_band(self):
        """Disable in-band remote management for the bridge.

        That configuration will apply to all controllers configured for the
        bridge.
        """
        other_config = {
            'disable-in-band': 'true'}
        self.ovsdb.db_set(
            'Bridge', self.br_name,
            ('other_config', other_config)).execute(check_error=True)

    def set_controller(self, controllers):
        self.ovsdb.set_controller(self.br_name,
                                  controllers).execute(check_error=True)
        # TODO(ralonsoh): this is a workaround for LP#1948642. When a new
        # controller is added, os-ken ``OfctlService`` needs to receive the
        # update event and register this new datapath.
        time.sleep(1)

    def del_controller(self):
        self.ovsdb.del_controller(self.br_name).execute(check_error=True)

    def get_controller(self):
        return self.ovsdb.get_controller(self.br_name).execute(
            check_error=True)

    def _set_bridge_fail_mode(self, mode):
        self.ovsdb.set_fail_mode(self.br_name, mode).execute(check_error=True)

    def set_secure_mode(self):
        self._set_bridge_fail_mode(FAILMODE_SECURE)

    def set_standalone_mode(self):
        self._set_bridge_fail_mode(FAILMODE_STANDALONE)

    def add_protocols(self, *protocols):
        existing_protocols = self.db_get_val(
            'Bridge', self.br_name, 'protocols')
        diff = set(protocols).difference(existing_protocols)
        if diff:
            self.ovsdb.db_add(
                'Bridge', self.br_name,
                'protocols', *diff).execute(check_error=True)
            # TODO(ralonsoh): this is a workaround for LP#1948642. When the OF
            # protocols are changed, the OF controller is restarted. This
            # sleep will provide time to os-ken ``OfctlService`` to receive
            # the update events of the restarted controllers and set them as
            # enabled.
            time.sleep(1)

    def use_at_least_protocol(self, protocol):
        """Calls to ovs-ofctl will use a protocol version >= 'protocol'"""
        self.add_protocols(protocol)
        self._highest_protocol_needed = max(self._highest_protocol_needed,
                                            protocol,
                                            key=version_from_protocol)
        self.initial_protocols.add(self._highest_protocol_needed)

    def set_igmp_snooping_state(self, state):
        state = bool(state)
        # NOTE(lucasagomes): The mcast-snooping-disable-flood-unregistered
        # has the opposite value of the config in Neutron. That's because
        # IGMP Neutron configs are more value consistent using True to
        # enable a feature and False to disable it.
        flood_value = ('false' if
            cfg.CONF.OVS.igmp_flood_unregistered else 'true')
        other_config = {
            'mcast-snooping-disable-flood-unregistered': flood_value}
        with self.ovsdb.transaction() as txn:
            txn.add(
                self.ovsdb.db_set('Bridge', self.br_name,
                                  ('mcast_snooping_enable', state)))
            txn.add(
                self.ovsdb.db_set('Bridge', self.br_name,
                                  ('other_config', other_config)))

    def set_igmp_snooping_flood(self, port_name):
        other_config = {
            'mcast-snooping-flood-reports': ovs_conf.get_igmp_flood_reports(),
            'mcast-snooping-flood': ovs_conf.get_igmp_flood()}
        self.ovsdb.db_set(
            'Port', port_name,
            ('other_config', other_config)).execute(
                check_error=True, log_errors=True)

    def create(self, secure_mode=False):
        other_config = {
            'mac-table-size': str(cfg.CONF.OVS.bridge_mac_table_size)}
        with self.ovsdb.transaction() as txn:
            txn.add(
                self.ovsdb.add_br(self.br_name,
                                  datapath_type=self.datapath_type))
            # the ovs-ofctl commands below in run_ofctl use OF10, so we
            # need to ensure that this version is enabled ; we could reuse
            # add_protocols, but doing ovsdb.db_add avoids doing two
            # transactions
            txn.add(
                self.ovsdb.db_add('Bridge', self.br_name,
                                  'protocols',
                                  *self.initial_protocols))
            txn.add(
                self.ovsdb.db_set('Bridge', self.br_name,
                                  ('other_config', other_config)))
            if secure_mode:
                txn.add(self.ovsdb.set_fail_mode(self.br_name,
                                                 FAILMODE_SECURE))

    def destroy(self):
        self.delete_bridge(self.br_name)

    def add_port(self, port_name, *interface_attr_tuples):
        with self.ovsdb.transaction() as txn:
            txn.add(self.ovsdb.add_port(self.br_name, port_name))
            if interface_attr_tuples:
                txn.add(self.ovsdb.db_set('Interface', port_name,
                                          *interface_attr_tuples))
        return self.get_port_ofport(port_name)

    def replace_port(self, port_name, *interface_attr_tuples):
        """Replace existing port or create it, and configure port interface."""

        # NOTE(xiaohhui): If del_port is inside the transaction, there will
        # only be one command for replace_port. This will cause the new port
        # not be found by system, which will lead to Bug #1519926.
        self.ovsdb.del_port(port_name).execute()
        with self.ovsdb.transaction() as txn:
            txn.add(self.ovsdb.add_port(self.br_name, port_name,
                                        may_exist=False))
            self._set_port_dead(port_name, txn)

            # TODO(mangelajo): We could accept attr tuples for the Port too
            # but, that could potentially break usage of this function in
            # stable branches (where we need to backport).
            # https://review.opendev.org/#/c/564825/4/neutron/agent/common/
            # ovs_lib.py@289
            if interface_attr_tuples:
                txn.add(self.ovsdb.db_set('Interface', port_name,
                                          *interface_attr_tuples))

    def _set_port_dead(self, port_name, txn):
        # NOTE(mangelajo): Port is added to dead vlan (4095) by default
        # until it's handled by the neutron-openvswitch-agent. Otherwise it
        # may trigger issues on ovs-vswitchd related to the
        # datapath flow revalidator thread, see lp#1767422
        txn.add(self.ovsdb.db_set(
            'Port', port_name, ('tag', ovs_constants.DEAD_VLAN_TAG)))
        # Just setting 'tag' to 4095 is not enough to prevent any traffic
        # to/from new port because "access" ports do not have 802.1Q header
        # and hence are not matched by default 4095-dropping rule.
        # So we also set "vlan_mode" attribute to "trunk" and "trunks"=[4095]
        # With this OVS normal pipeline will allow only packets tagged with
        # 4095 from such ports, which normally not happens,
        # but even if it does - default rule in br-int will drop them anyway.
        # Thus untagged packets from such ports will also be dropped until
        # ovs agent sets proper VLAN tag and clears vlan_mode to default
        # ("access"). See lp#1930414 for details.
        txn.add(self.ovsdb.db_set(
            'Port', port_name, ('vlan_mode', 'trunk')))
        txn.add(self.ovsdb.db_set(
            'Port', port_name, ('trunks', ovs_constants.DEAD_VLAN_TAG)))

    def delete_port(self, port_name):
        self.ovsdb.del_port(port_name, self.br_name).execute()

    def run_ofctl(self, cmd, args, process_input=None):
        full_args = ["ovs-ofctl", cmd,
                     "-O", self._highest_protocol_needed,
                     self.br_name] + args
        # TODO(kevinbenton): This error handling is really brittle and only
        # detects one specific type of failure. The callers of this need to
        # be refactored to expect errors so we can re-raise and they can
        # take appropriate action based on the type of error.
        for i in range(1, 11):
            try:
                return utils.execute(full_args, run_as_root=True,
                                     process_input=process_input,
                                     privsep_exec=True)
            except Exception as e:
                if "failed to connect to socket" in str(e):
                    LOG.debug("Failed to connect to OVS. Retrying "
                              "in 1 second. Attempt: %s/10", i)
                    time.sleep(1)
                    continue
                LOG.error("Unable to execute %(cmd)s. Exception: "
                          "%(exception)s",
                          {'cmd': full_args, 'exception': e})
                break

    def count_flows(self):
        flow_list = self.run_ofctl("dump-flows", []).split("\n")[1:]
        return len(flow_list) - 1

    def remove_all_flows(self):
        self.run_ofctl("del-flows", [])

    def list_meter_features(self):
        # For fullstack test mainly
        f_list = self.run_ofctl("meter-features", []).split("\n")[1:]
        max_meter = max_bands = support_drop = support_caps = None
        for output in f_list:
            match = MAX_METER_REGEX.match(output)
            if match:
                max_meter = match.group(1)
                max_bands = match.group(2)
            match = BAND_TYPES_REGEX.match(output)
            if match:
                support_drop = match.group(1)
            match = CAPS_REGEX.match(output)
            if match:
                support_caps = [match.group(1), match.group(2),
                                match.group(3), match.group(4)]
        return all([max_meter, max_bands, support_drop, support_caps])

    @_ovsdb_retry
    def _get_port_val(self, port_name, port_val):
        return self.db_get_val("Interface", port_name, port_val)

    def get_port_ofport(self, port_name):
        """Get the port's assigned ofport, retrying if not yet assigned."""
        ofport = INVALID_OFPORT
        try:
            ofport = self._get_port_val(port_name, "ofport")
        except tenacity.RetryError:
            LOG.exception("Timed out retrieving ofport on port %s.",
                          port_name)
        return ofport

    @_ovsdb_retry
    def _get_datapath_id(self):
        return self.db_get_val('Bridge', self.br_name, 'datapath_id')

    def get_datapath_id(self):
        try:
            return self._get_datapath_id()
        except tenacity.RetryError:
            # if ovs fails to find datapath_id then something is likely to be
            # broken here
            LOG.exception("Timed out retrieving datapath_id on bridge %s.",
                          self.br_name)
            raise RuntimeError(_('No datapath_id on bridge %s') % self.br_name)

    def do_action_flows_by_group_id(self, action, flows_by_group_id,
                                    use_bundle=False):
        if self._flows_per_port:
            # Group flow actions per port.
            for flow_group_id, flows in flows_by_group_id.items():
                self.do_action_flows(action, flows, use_bundle=use_bundle,
                                     flow_group_id=flow_group_id)
        else:
            # Group all actions in one single list without any group ID
            # reference.
            flows = [item for _list in flows_by_group_id.values()
                     for item in _list]
            self.do_action_flows(action, flows, use_bundle=use_bundle)

    def do_action_flows(self, action, kwargs_list, use_bundle=False,
                        flow_group_id=None):
        # we can't mix strict and non-strict, so we'll use the first kw
        # and check against other kw being different
        strict = kwargs_list[0].get('strict', False)

        for kw in kwargs_list:
            if action == 'del':
                if kw.get('cookie') == COOKIE_ANY:
                    # special value COOKIE_ANY was provided, unset
                    # cookie to match flows whatever their cookie is
                    kw.pop('cookie')
                    if kw.get('cookie_mask'):  # non-zero cookie mask
                        raise Exception(_("cookie=COOKIE_ANY but cookie_mask "
                                          "set to %s") % kw.get('cookie_mask'))
                elif 'cookie' in kw:
                    # a cookie was specified, use it
                    kw['cookie'] = check_cookie_mask(kw['cookie'])
                else:
                    # nothing was specified about cookies, use default
                    kw['cookie'] = "%d/-1" % self._default_cookie
            else:
                if 'cookie' not in kw:
                    kw['cookie'] = self._default_cookie

            if action in ('mod', 'del'):
                if kw.pop('strict', False) != strict:
                    msg = ("cannot mix 'strict' and not 'strict' in a batch "
                           "call")
                    raise exceptions.InvalidInput(error_message=msg)
            else:
                if kw.pop('strict', False):
                    msg = "cannot use 'strict' with 'add' action"
                    raise exceptions.InvalidInput(error_message=msg)

        extra_param = ["--strict"] if strict else []

        if action == 'del' and {} in kwargs_list:
            # the 'del' case simplifies itself if kwargs_list has at least
            # one item that matches everything
            self.run_ofctl('%s-flows' % action, [])
        else:
            flow_strs = [_build_flow_expr_str(kw, action, strict)
                         for kw in kwargs_list]
            LOG.debug("Processing %d OpenFlow rules.", len(flow_strs))
            if use_bundle:
                extra_param.append('--bundle')

            if flow_group_id:
                # NOTE(ralonsoh): all flows belonging to a port will be written
                # atomically in the same command.
                step = len(flow_strs)
            else:
                # No group ID defined (flows are not grouped per port). Use the
                # default batch step value "openflow_number_processing_step".
                step = common_constants.AGENT_RES_PROCESSING_STEP

            for i in range(0, len(flow_strs), step):
                self.run_ofctl('%s-flows' % action, extra_param + ['-'],
                               '\n'.join(flow_strs[i:i + step]))

    def add_flow(self, **kwargs):
        self.do_action_flows('add', [kwargs])

    def mod_flow(self, **kwargs):
        self.do_action_flows('mod', [kwargs])

    def delete_flows(self, **kwargs):
        self.do_action_flows('del', [kwargs])

    def dump_flows_for_table(self, table):
        return self.dump_flows_for(table=table)

    def dump_flows_for(self, **kwargs):
        retval = None
        if "cookie" in kwargs:
            kwargs["cookie"] = check_cookie_mask(str(kwargs["cookie"]))
        flow_str = ",".join("=".join([key, str(val)])
                            for key, val in kwargs.items())

        flows = self.run_ofctl("dump-flows", [flow_str])
        if flows:
            retval = '\n'.join(item for item in flows.splitlines()
                               if is_a_flow_line(item))
        return retval

    def dump_all_flows(self):
        return [f for f in self.run_ofctl("dump-flows", []).splitlines()
                if is_a_flow_line(f)]

    def deferred(self, *args, **kwargs):
        return DeferredOVSBridge(self, *args, **kwargs)

    def add_tunnel_port(self, port_name, remote_ip, local_ip,
                        tunnel_type=p_const.TYPE_GRE,
                        vxlan_udp_port=p_const.VXLAN_UDP_PORT,
                        dont_fragment=True,
                        tunnel_csum=False,
                        tos=None):
        if tunnel_type == p_const.TYPE_GRE:
            tunnel_type = get_gre_tunnel_port_type(remote_ip, local_ip)
        attrs = [('type', tunnel_type)]
        # TODO(twilson) This is an OrderedDict solely to make a test happy
        options = collections.OrderedDict()
        vxlan_uses_custom_udp_port = (
            tunnel_type == p_const.TYPE_VXLAN and
            vxlan_udp_port != p_const.VXLAN_UDP_PORT
        )
        if vxlan_uses_custom_udp_port:
            options['dst_port'] = str(vxlan_udp_port)
        options['df_default'] = str(dont_fragment).lower()
        options['remote_ip'] = remote_ip
        options['local_ip'] = local_ip
        options['in_key'] = 'flow'
        options['out_key'] = 'flow'
        # NOTE(moshele): pkt_mark is not supported when using ovs hw-offload,
        # therefore avoid clear mark on encapsulating packets when it's
        # enabled
        if not self.is_hw_offload_enabled:
            options['egress_pkt_mark'] = '0'
        if tunnel_csum:
            options['csum'] = str(tunnel_csum).lower()
        if tos:
            options['tos'] = str(tos)
        if tunnel_type == TYPE_GRE_IP6:
            # NOTE(slaweq) According to the OVS documentation L3 GRE tunnels
            # over IPv6 are not supported.
            options['packet_type'] = 'legacy_l2'
        attrs.append(('options', options))

        return self.add_port(port_name, *attrs)

    def add_patch_port(self, local_name, remote_name):
        attrs = [('type', 'patch'),
                 ('options', {'peer': remote_name})]
        return self.add_port(local_name, *attrs)

    def get_iface_name_list(self):
        # get the interface name list for this bridge
        return self.ovsdb.list_ifaces(self.br_name).execute(check_error=True)

    def get_port_name_list(self):
        # get the port name list for this bridge
        return self.ovsdb.list_ports(self.br_name).execute(check_error=True)

    def get_port_stats(self, port_name):
        return self.db_get_val("Interface", port_name, "statistics")

    def get_ports_attributes(self, table, columns=None, ports=None,
                             check_error=True, log_errors=True,
                             if_exists=False):
        port_names = ports or self.get_port_name_list()
        if not port_names:
            return []
        return (self.ovsdb.db_list(table, port_names, columns=columns,
                                   if_exists=if_exists).
                execute(check_error=check_error, log_errors=log_errors))

    # returns a VIF object for each VIF port
    def get_vif_ports(self, ofport_filter=None):
        edge_ports = []
        port_info = self.get_ports_attributes(
            'Interface', columns=['name', 'external_ids', 'ofport'],
            if_exists=True)
        for port in port_info:
            name = port['name']
            external_ids = port['external_ids']
            ofport = port['ofport']
            if ofport_filter and ofport in ofport_filter:
                continue
            if "iface-id" in external_ids and "attached-mac" in external_ids:
                p = VifPort(name, ofport, external_ids["iface-id"],
                            external_ids["attached-mac"], self)
                edge_ports.append(p)

        return edge_ports

    def get_vif_port_to_ofport_map(self):
        results = self.get_ports_attributes(
            'Interface', columns=['name', 'external_ids', 'ofport'],
            if_exists=True)
        port_map = {}
        for r in results:
            # fall back to basic interface name
            key = self.portid_from_external_ids(r['external_ids']) or r['name']
            try:
                port_map[key] = int(r['ofport'])
            except TypeError:
                # port doesn't yet have an ofport entry so we ignore it
                pass
        return port_map

    def get_vif_port_set(self):
        edge_ports = set()
        results = self.get_ports_attributes(
            'Interface', columns=['name', 'external_ids', 'ofport'],
            if_exists=True)
        for result in results:
            if result['ofport'] == UNASSIGNED_OFPORT:
                LOG.warning("Found not yet ready openvswitch port: %s",
                            result['name'])
            elif result['ofport'] == INVALID_OFPORT:
                LOG.warning("Found failed openvswitch port: %s",
                            result['name'])
            elif 'attached-mac' in result['external_ids']:
                port_id = self.portid_from_external_ids(result['external_ids'])
                if port_id:
                    edge_ports.add(port_id)
        return edge_ports

    def portid_from_external_ids(self, external_ids):
        if 'iface-id' in external_ids:
            return external_ids['iface-id']

    def get_port_tag_dict(self):
        """Get a dict of port names and associated vlan tags.

        e.g. the returned dict is of the following form::

            {u'int-br-eth2': [],
             u'patch-tun': [],
             u'qr-76d9e6b6-21': 1,
             u'tapce5318ff-78': 1,
             u'tape1400310-e6': 1}

        The TAG ID is only available in the "Port" table and is not available
        in the "Interface" table queried by the get_vif_port_set() method.

        """
        results = self.get_ports_attributes(
            'Port', columns=['name', 'tag'], if_exists=True)
        return {p['name']: p['tag'] for p in results}

    def get_vifs_by_ids(self, port_ids):
        interface_info = self.get_ports_attributes(
            "Interface", columns=["name", "external_ids", "ofport"],
            if_exists=True)
        by_id = {x['external_ids'].get('iface-id'): x for x in interface_info}
        result = {}
        for port_id in port_ids:
            result[port_id] = None
            if port_id not in by_id:
                LOG.info("Port %(port_id)s not present in bridge "
                         "%(br_name)s",
                         {'port_id': port_id, 'br_name': self.br_name})
                continue
            pinfo = by_id[port_id]
            mac = pinfo['external_ids'].get('attached-mac')
            result[port_id] = VifPort(pinfo['name'], pinfo['ofport'],
                                      port_id, mac, self)
        return result

    def get_vif_port_by_id(self, port_id):
        ports = self.ovsdb.db_find(
            'Interface', ('external_ids', '=', {'iface-id': port_id}),
            ('external_ids', '!=', {'attached-mac': ''}),
            columns=['external_ids', 'name', 'ofport']).execute()
        for port in ports:
            if self.br_name != self.get_bridge_for_iface(port['name']):
                continue
            mac = port['external_ids'].get('attached-mac')
            return VifPort(port['name'], port['ofport'], port_id, mac, self)
        LOG.info("Port %(port_id)s not present in bridge %(br_name)s",
                 {'port_id': port_id, 'br_name': self.br_name})

    def get_bridge_patch_ports_ofports(self):
        ports = self.ovsdb.db_find(
            'Interface', ('type', '=', 'patch'),
            columns=['name', 'ofport']).execute()
        patch_ports = []
        for port in ports:
            if self.br_name != self.get_bridge_for_iface(port['name']):
                continue
            patch_ports.append(port['ofport'])
        return patch_ports

    def delete_ports(self, all_ports=False):
        if all_ports:
            port_names = self.get_port_name_list()
        else:
            port_names = (port.port_name for port in self.get_vif_ports())

        for port_name in port_names:
            self.delete_port(port_name)

    def get_local_port_mac(self):
        """Retrieve the mac of the bridge's local port."""
        address = ip_lib.IPDevice(self.br_name).link.address
        if address:
            return address
        else:
            msg = _('Unable to determine mac address for %s') % self.br_name
            raise Exception(msg)

    def set_controllers_inactivity_probe(self, interval):
        """Set bridge controllers inactivity probe interval.

        :param interval: inactivity_probe value in seconds.
        """
        self.set_controller_field('inactivity_probe', interval * 1000)

    def _set_egress_bw_limit_for_port(self, port_name, max_kbps,
                                      max_burst_kbps, check_error=True):
        with self.ovsdb.transaction(check_error=check_error) as txn:
            txn.add(self.ovsdb.db_set('Interface', port_name,
                                      ('ingress_policing_rate', max_kbps)))
            txn.add(self.ovsdb.db_set('Interface', port_name,
                                      ('ingress_policing_burst',
                                       max_burst_kbps)))

    def create_egress_bw_limit_for_port(self, port_name, max_kbps,
                                        max_burst_kbps):
        self._set_egress_bw_limit_for_port(
            port_name, max_kbps, max_burst_kbps)

    def get_egress_bw_limit_for_port(self, port_name):

        max_kbps = self.db_get_val('Interface', port_name,
                                   'ingress_policing_rate')
        max_burst_kbps = self.db_get_val('Interface', port_name,
                                         'ingress_policing_burst')

        max_kbps = max_kbps or None
        max_burst_kbps = max_burst_kbps or None

        return max_kbps, max_burst_kbps

    def delete_egress_bw_limit_for_port(self, port_name):
        if not self.port_exists(port_name):
            return
        self._set_egress_bw_limit_for_port(port_name, 0, 0, check_error=False)

    def _update_bw_limit_profile_dpdk(self, txn, port_name, qos_uuid,
                                      other_config):
        if qos_uuid:
            txn.add(self.ovsdb.db_set(
                'QoS', qos_uuid, ('other_config', other_config)))
        else:
            external_ids = {'id': port_name}
            qos_uuid = txn.add(
                self.ovsdb.db_create(
                    'QoS', external_ids=external_ids, type='egress-policer',
                    other_config=other_config))
        return qos_uuid

    def set_queue_for_ingress_bandwidth_limit(self):
        # reg3 is used to memoize if queue was set or not. If it is first visit
        # to table 0 for a packet (i.e. reg3 == 0), set queue and memoize (i.e.
        # load 1 to reg3), then goto table 0 again. The packet will be handled
        # as usual when the second visit to table 0.
        # For min bw reg4 is used for the same purpose. In case if there we
        # would need one of those registries for something else in the future
        # we can try to use same reg4 for both OF rules, this one and the one
        # which sets pkt_mark for minimum bandwidth and play with bitmask
        self.add_flow(
            table=ovs_constants.LOCAL_SWITCHING,
            reg3=0,
            priority=200,
            actions=("set_queue:%s,load:1->NXM_NX_REG3[0],"
                     "resubmit(,%s)" % (QOS_DEFAULT_QUEUE,
                                        ovs_constants.LOCAL_SWITCHING)))

    def _update_ingress_bw_limit_for_port(
            self, port_name, max_kbps, max_burst_kbps):
        queue_id = self._update_queue(
            port_name, QOS_DEFAULT_QUEUE,
            qos_constants.RULE_TYPE_BANDWIDTH_LIMIT,
            max_kbps=max_kbps, max_burst_kbps=max_burst_kbps)
        qos_id, qos_queues = self._find_qos(
            port_name,
            qos_constants.RULE_TYPE_BANDWIDTH_LIMIT)
        if qos_queues:
            qos_queues[QOS_DEFAULT_QUEUE] = queue_id
        else:
            qos_queues = {QOS_DEFAULT_QUEUE: queue_id}
        qos_id = self._update_qos(
            port_name,
            qos_constants.RULE_TYPE_BANDWIDTH_LIMIT,
            qos_id=qos_id, queues=qos_queues)
        self._set_port_qos(port_name, qos_id=qos_id)

    def _update_ingress_bw_limit_for_dpdk_port(
            self, port_name, max_kbps, max_burst_kbps):
        # cir and cbs should be set in bytes instead of bits
        max_bw_in_bits = max_kbps * p_const.SI_BASE
        max_burst_in_bits = max_burst_kbps * p_const.SI_BASE
        qos_other_config = {
            'cir': str(max_bw_in_bits // 8),
            'cbs': str(max_burst_in_bits // 8)
        }
        qos_id, qos_queues = self._find_qos(port_name)
        with self.ovsdb.transaction(check_error=True) as txn:
            qos_uuid = self._update_bw_limit_profile_dpdk(
                txn, port_name, qos_id, qos_other_config)
            txn.add(self.ovsdb.db_set(
                'Port', port_name, ('qos', qos_uuid)))

    def update_ingress_bw_limit_for_port(self, port_name, max_kbps,
                                         max_burst_kbps):
        port_type = self._get_port_val(port_name, "type")
        if port_type in ovs_constants.OVS_DPDK_PORT_TYPES:
            self._update_ingress_bw_limit_for_dpdk_port(
                port_name, max_kbps, max_burst_kbps)
        else:
            self._update_ingress_bw_limit_for_port(
                port_name, max_kbps, max_burst_kbps)

    def get_ingress_bw_limit_for_port(self, port_name):
        qos_max_kbps = None
        max_burst_kbit = None

        queue = self._find_queue(
            port_name, _type=qos_constants.RULE_TYPE_BANDWIDTH_LIMIT)
        if queue:
            other_config = queue['other_config']
            max_bw_in_bits = other_config.get('max-rate')
            qos_max_kbps = int(int(max_bw_in_bits) / p_const.SI_BASE)
            max_burst_in_bits = other_config.get('burst')
            if max_burst_in_bits is not None:
                max_burst_kbit = int(int(max_burst_in_bits) / p_const.SI_BASE)

        return qos_max_kbps, max_burst_kbit

    def delete_ingress_bw_limit_for_port(self, port_name):
        qos_id, qos_queues = self._find_qos(
            port_name,
            qos_constants.RULE_TYPE_BANDWIDTH_LIMIT)
        if not qos_queues:
            return
        if QOS_DEFAULT_QUEUE in qos_queues.keys():
            queue_uuid = qos_queues.pop(QOS_DEFAULT_QUEUE)
            if qos_queues:
                self._update_qos(
                    port_name,
                    qos_constants.RULE_TYPE_BANDWIDTH_LIMIT,
                    qos_id=qos_id, queues=qos_queues)
            self.ovsdb.db_clear('Port', port_name, 'qos').execute(
                check_error=False)
            if not qos_queues:
                self._delete_qos(qos_id)
            self._delete_queue(
                queue_uuid, qos_constants.RULE_TYPE_BANDWIDTH_LIMIT)

    def set_controller_field(self, field, value):
        attr = [(field, value)]
        controllers = self.db_get_val('Bridge', self.br_name, 'controller')
        controllers = [controllers] if isinstance(
            controllers, uuid.UUID) else controllers
        with self.ovsdb.transaction(check_error=True) as txn:
            for controller_uuid in controllers:
                txn.add(self.ovsdb.db_set(
                    'Controller', controller_uuid, *attr))

    def set_controller_rate_limit(self, controller_rate_limit):
        """Set bridge controller_rate_limit

        :param controller_rate_limit: at least 100
        """
        if controller_rate_limit < CTRL_RATE_LIMIT_MIN:
            LOG.info("rate limit's value must be at least 100")
            controller_rate_limit = CTRL_RATE_LIMIT_MIN
        self.set_controller_field(
            'controller_rate_limit', controller_rate_limit)

    def set_controller_burst_limit(self, controller_burst_limit):
        """Set bridge controller_burst_limit

        :param controller_burst_limit: at least 25
        """
        if controller_burst_limit < CTRL_BURST_LIMIT_MIN:
            LOG.info("burst limit's value must be at least 25")
            controller_burst_limit = CTRL_BURST_LIMIT_MIN
        self.set_controller_field(
            'controller_burst_limit', controller_burst_limit)

    def set_datapath_id(self, datapath_id):
        dpid_cfg = {'datapath-id': datapath_id}
        self.set_db_attribute('Bridge', self.br_name, 'other_config', dpid_cfg,
                              check_error=True)

    def get_egress_min_bw_for_port(self, port_id):
        queue = self._find_queue(port_id)
        if not queue:
            return

        min_bps = queue['other_config'].get('min-rate')
        return int(int(min_bps) / 1000) if min_bps else None

    def _set_pkt_mark_for_minimum_bandwidth(self, queue_num):
        # reg4 is used to memoize if queue was set or not. If it is first visit
        # to table 0 for a packet (i.e. reg4 == 0), set queue and memoize (i.e.
        # load 1 to reg4), then goto table 0 again. The packet will be handled
        # as usual when the second visit to table 0.
        self.add_flow(
            table=ovs_constants.LOCAL_SWITCHING,
            in_port=queue_num,
            reg4=0,
            priority=200,
            actions=("set_field:%s->pkt_mark,load:1->NXM_NX_REG4[0],"
                     "resubmit(,%s)" % (queue_num,
                                        ovs_constants.LOCAL_SWITCHING)))

    def set_queue_for_minimum_bandwidth(self, queue_num):
        # reg4 is used to memoize if queue was set or not. If it is first visit
        # to table 0 for a packet (i.e. reg4 == 0), set queue and memoize (i.e.
        # load 1 to reg4), then goto table 0 again. The packet will be handled
        # as usual when the second visit to table 0.
        patch_ports = self.get_bridge_patch_ports_ofports()
        for patch_port in patch_ports:
            self.add_flow(
                table=0,
                in_port=patch_port,
                pkt_mark=queue_num,
                reg4=0,
                priority=200,
                actions=("set_queue:%s,load:1->NXM_NX_REG4[0],"
                         "resubmit(,0)" % queue_num))

    def _unset_pkt_mark_for_minimum_bandwidth(self, queue_num):
        self.delete_flows(
            table=ovs_constants.LOCAL_SWITCHING,
            in_port=queue_num,
            reg4=0)

    def update_minimum_bandwidth_queue(self, port_id, egress_port_names,
                                       queue_num, min_kbps):
        queue_num = int(queue_num)
        queue_id = self._update_queue(
            port_id, queue_num, qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH,
            min_kbps=min_kbps)
        qos_id, qos_queues = self._find_qos(
            self._min_bw_qos_id,
            qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH)
        if qos_queues:
            qos_queues[queue_num] = queue_id
        else:
            qos_queues = {queue_num: queue_id}
        qos_id = self._update_qos(
            self._min_bw_qos_id,
            qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH,
            qos_id=qos_id, queues=qos_queues)
        for egress_port_name in egress_port_names:
            self._set_port_qos(egress_port_name, qos_id=qos_id)
        self._set_pkt_mark_for_minimum_bandwidth(queue_num)
        return qos_id

    def delete_minimum_bandwidth_queue(self, port_id):
        queue = self._find_queue(port_id)
        if not queue:
            return
        queue_num = int(queue['external_ids']['queue-num'])
        self._unset_pkt_mark_for_minimum_bandwidth(queue_num)
        qos_id, qos_queues = self._find_qos(
            self._min_bw_qos_id,
            qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH)
        if not qos_queues:
            return
        if queue_num in qos_queues.keys():
            qos_queues.pop(queue_num)
            if qos_queues:
                self._update_qos(
                    self._min_bw_qos_id,
                    qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH,
                    qos_id=qos_id, queues=qos_queues)
            else:
                # Find the physical bridge interface with the QoS assigned and
                # unset it.
                for port in self.ovsdb.db_find(
                        'Port', ('qos', '=', qos_id),
                        columns=['_uuid', 'qos']).execute(check_error=True):
                    self.ovsdb.db_clear('Port', port['_uuid'],
                                        'qos').execute(check_error=True)
                self._delete_qos(qos_id)
            self._delete_queue(
                queue['_uuid'], qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH)

    def clear_bandwidth_qos(self):
        qoses = []
        qos_types = [
            (self._min_bw_qos_id,
             qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH),
            (None,
             qos_constants.RULE_TYPE_BANDWIDTH_LIMIT)]
        for rule_type_id, qos_type in qos_types:
            qoses += self._list_qos(_id=rule_type_id, qos_type=qos_type)

        for qos in qoses:
            qos_id = qos['_uuid']
            queues = {num: queue.uuid
                      for num, queue in qos['queues'].items()}
            ports = self.ovsdb.db_find(
                'Port',
                ('qos', '=', qos_id),
                colmuns=['name']).execute(check_error=True)
            for port in ports:
                self._set_port_qos(port['name'])
            self._delete_qos(qos_id)
            for queue_uuid in queues.values():
                self._delete_queue(queue_uuid)

    def _update_queue(self, port_id, queue_num, queue_type, max_kbps=None,
                      max_burst_kbps=None, min_kbps=None):
        queue = self._find_queue(port_id, _type=queue_type)
        other_config = dict(queue['other_config']) if queue else {}

        if max_kbps:
            other_config['max-rate'] = str(int(max_kbps) * p_const.SI_BASE)
            other_config['burst'] = str(int(max_burst_kbps) * p_const.SI_BASE)
        if min_kbps:
            other_config['min-rate'] = str(min_kbps * p_const.SI_BASE)
            if 'max-rate' not in other_config:
                other_config['max-rate'] = str(OVS_MAX_RATE)

        if queue and queue['_uuid']:
            if queue['other_config'] != other_config:
                self.set_db_attribute('Queue', queue['_uuid'], 'other_config',
                                      other_config, check_error=True)
        else:
            # NOTE(ralonsoh): "external_ids" is a map of string-string pairs
            external_ids = {
                'port': str(port_id),
                'type': str(queue_type),
                'queue-num': str(queue_num)}
            self.ovsdb.db_create(
                'Queue', other_config=other_config,
                external_ids=external_ids).execute(check_error=True)
            queue = self._find_queue(port_id, _type=queue_type)
        return queue['_uuid']

    def _find_queue(self, port_id, _type=None):
        # NOTE(ralonsoh): in ovsdb native library, '{>=}' operator is not
        # implemented yet. This is a workaround: list all queues and compare
        # the external_ids key needed.
        _type = _type or qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH
        queues = self._list_queues(port=port_id, _type=_type)
        if queues:
            return queues[0]
        return None

    def _list_queues(self, _type=None, port=None):
        queues = self.ovsdb.db_list(
            'Queue',
            columns=['_uuid', 'external_ids', 'other_config']).execute(
                check_error=True)
        if port:
            queues = [queue for queue in queues
                      if queue['external_ids'].get('port') == str(port)]
        if _type:
            queues = [queue for queue in queues
                      if queue['external_ids'].get('type') == str(_type)]
        return queues

    def _delete_qos(self, qos_id):
        try:
            self.ovsdb.db_destroy('QoS', qos_id).execute(check_error=True)
        except idlutils.RowNotFound:
            LOG.info('OVS QoS %s was already deleted', str(qos_id))

    def _delete_queue(self, queue_id, qos_type=None):
        try:
            self.ovsdb.db_destroy('Queue', queue_id).execute(
                check_error=True)
        except idlutils.RowNotFound:
            LOG.info('OVS Queue %s was already deleted', queue_id)
        except RuntimeError as exc:
            with excutils.save_and_reraise_exception():
                if 'referential integrity violation' not in str(exc):
                    return
                qos_regs = self._list_qos(qos_type=qos_type)
                qos_uuids = []
                for qos_reg in qos_regs:
                    queue_nums = [num for num, q in qos_reg['queues'].items()
                                  if q.uuid == queue_id]
                    if queue_nums:
                        qos_uuids.append(str(qos_reg['_uuid']))
                LOG.error('Queue %(queue)s was still in use by the following '
                          'QoS rules: %(qoses)s',
                          {'queue': str(queue_id),
                           'qoses': ', '.join(sorted(qos_uuids))})

    def _update_qos(self, rule_type_id, rule_type, qos_id=None, queues=None):
        queues = queues or {}
        if not qos_id:
            external_ids = {'id': rule_type_id,
                            '_type': rule_type}
            other_config = {'max-rate': str(OVS_MAX_RATE)}
            self.ovsdb.db_create(
                'QoS',
                type='linux-htb',
                queues=queues,
                external_ids=external_ids,
                other_config=other_config).execute(check_error=True)
            qos_id, _ = self._find_qos(rule_type_id, rule_type)
        else:
            self.clear_db_attribute('QoS', qos_id, 'queues')
            if queues:
                self.set_db_attribute('QoS', qos_id, 'queues', queues,
                                      check_error=True)
        return qos_id

    def _list_qos(self, _id=None, qos_type=None):
        external_ids = {}
        if _id:
            external_ids['id'] = _id
        if qos_type:
            external_ids['_type'] = qos_type
        if external_ids:
            return self.ovsdb.db_find(
                'QoS',
                ('external_ids', '=', external_ids),
                colmuns=['_uuid', 'queues']).execute(check_error=True)

        return self.ovsdb.db_find(
            'QoS', colmuns=['_uuid', 'queues']).execute(check_error=True)

    def _find_qos(self, rule_type_id, qos_type=None):
        qos_regs = self._list_qos(_id=rule_type_id, qos_type=qos_type)
        if qos_regs:
            queues = {num: queue.uuid
                      for num, queue in qos_regs[0]['queues'].items()}
            return qos_regs[0]['_uuid'], queues
        return None, None

    def _set_port_qos(self, port_name, qos_id=None):
        if qos_id:
            self.set_db_attribute('Port', port_name, 'qos', qos_id,
                                  check_error=True)
        else:
            self.clear_db_attribute('Port', port_name, 'qos')

    def get_bridge_ports(self, port_type=None):
        port_names = self.get_port_name_list() + [self.br_name]
        ports = self.get_ports_attributes('Interface',
                                          ports=port_names,
                                          columns=['name', 'type'],
                                          if_exists=True) or []
        if port_type is None:
            return ports
        elif not isinstance(port_type, list):
            port_type = [port_type]
        return [port['name'] for port in ports if port['type'] in port_type]

    def get_port_tag_by_name(self, port_name):
        # At the very beginning of port processing, the port tag
        # may not set to ovsdb Port. But, we set the tag to
        # other_config.
        return self.get_value_from_other_config(port_name, 'tag', int)

    def get_value_from_other_config(self, port_name,
                                    key, value_type=None):
        try:
            other_config = self.db_get_val(
                'Port', port_name, 'other_config') or {}
            value = other_config.get(key)
            if value is not None:
                if value_type:
                    return value_type(value)
                return value
        except (TypeError, ValueError):
            raise ovs_exc.OVSDBPortError(port=port_name)

    def set_value_to_other_config(self, port_name, key, value):
        other_config = self.db_get_val(
            'Port', port_name, 'other_config')
        if isinstance(other_config, dict):
            other_config[key] = str(value)
            # set_db_attribute does not work
            with self.ovsdb.transaction() as txn:
                txn.add(
                    self.ovsdb.db_set('Port', port_name,
                                      ('other_config', other_config)))

    def remove_value_from_other_config(self, port_name, key):
        other_config = self.db_get_val(
            'Port', port_name, 'other_config')
        if isinstance(other_config, dict):
            other_config.pop(key, None)
            # set_db_attribute does not work
            with self.ovsdb.transaction() as txn:
                txn.add(self.ovsdb.db_clear('Port', port_name, "other_config"))
                txn.add(self.ovsdb.db_set('Port', port_name,
                                          ('other_config', other_config)))

    def __enter__(self):
        self.create()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.destroy()


class DeferredOVSBridge(object):
    '''Deferred OVSBridge.

    This class wraps add_flow, mod_flow and delete_flows calls to an OVSBridge
    and defers their application until apply_flows call in order to perform
    bulk calls. It wraps also ALLOWED_PASSTHROUGHS calls to avoid mixing
    OVSBridge and DeferredOVSBridge uses.
    This class can be used as a context, in such case apply_flows is called on
    __exit__ except if an exception is raised.
    This class is not thread-safe, that's why for every use a new instance
    must be implemented.
    '''
    ALLOWED_PASSTHROUGHS = 'add_port', 'add_tunnel_port', 'delete_port'

    def __init__(self, br, full_ordered=False,
                 order=('add', 'mod', 'del'), use_bundle=False):
        '''Constructor.

        :param br: wrapped bridge
        :param full_ordered: Optional, disable flow reordering (slower)
        :param order: Optional, define in which order flow are applied
        :param use_bundle: Optional, a bool whether --bundle should be passed
                           to all ofctl commands. Default is set to False.
        '''

        self.br = br
        self.full_ordered = full_ordered
        self.order = order
        if not self.full_ordered:
            self.weights = dict((y, x) for x, y in enumerate(self.order))
        self.action_flow_tuples = []
        self.use_bundle = use_bundle

    def __getattr__(self, name):
        if name in self.ALLOWED_PASSTHROUGHS:
            return getattr(self.br, name)
        raise AttributeError(name)

    def add_flow(self, flow_group_id=None, **kwargs):
        self.action_flow_tuples.append(
            ActionFlowTuple('add', kwargs, flow_group_id))

    def mod_flow(self, **kwargs):
        self.action_flow_tuples.append(ActionFlowTuple('mod', kwargs, None))

    def delete_flows(self, **kwargs):
        self.action_flow_tuples.append(ActionFlowTuple('del', kwargs, None))

    def apply_flows(self):
        action_flow_tuples = self.action_flow_tuples
        self.action_flow_tuples = []
        if not action_flow_tuples:
            return

        if not self.full_ordered:
            action_flow_tuples.sort(key=lambda flow: self.weights[flow.action])

        flows_by_action = itertools.groupby(action_flow_tuples,
                                            key=lambda af: af.action)
        for action, flows in flows_by_action:
            flows_by_group_id = collections.defaultdict(list)
            for flow in flows:
                flows_by_group_id[flow.flow_group_id].append(flow.flow)
            self.br.do_action_flows_by_group_id(action, flows_by_group_id,
                                                self.use_bundle)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.apply_flows()
        else:
            LOG.exception("OVS flows could not be applied on bridge %s",
                          self.br.br_name)


def _build_flow_expr_str(flow_dict, cmd, strict):
    flow_expr_arr = []
    actions = None

    if cmd == 'add':
        flow_expr_arr.append("hard_timeout=%s" %
                             flow_dict.pop('hard_timeout', '0'))
        flow_expr_arr.append("idle_timeout=%s" %
                             flow_dict.pop('idle_timeout', '0'))
        flow_expr_arr.append("priority=%s" %
                             flow_dict.pop('priority', '1'))
    elif 'priority' in flow_dict:
        if not strict:
            msg = _("Cannot match priority on flow deletion or modification "
                    "without 'strict'")
            raise exceptions.InvalidInput(error_message=msg)

    if cmd != 'del':
        if "actions" not in flow_dict:
            msg = _("Must specify one or more actions on flow addition"
                    " or modification")
            raise exceptions.InvalidInput(error_message=msg)
        actions = "actions=%s" % flow_dict.pop('actions')

    for key, value in flow_dict.items():
        if key == 'proto':
            flow_expr_arr.append(value)
        else:
            flow_expr_arr.append("%s=%s" % (key, str(value)))

    if actions:
        flow_expr_arr.append(actions)

    return ','.join(flow_expr_arr)


def generate_random_cookie():
    # The OpenFlow spec forbids use of -1
    return random.randrange(UINT64_BITMASK)


def check_cookie_mask(cookie):
    cookie = str(cookie)
    if '/' not in cookie:
        return cookie + '/-1'
    else:
        return cookie


def is_a_flow_line(line):
    # this is used to filter out from ovs-ofctl dump-flows the lines that
    # are not flow descriptions but mere indications of the type of openflow
    # message that was used ; e.g.:
    #
    # # ovs-ofctl dump-flows br-int
    # NXST_FLOW reply (xid=0x4):
    #  cookie=0xb7dff131a697c6a5, duration=2411726.809s, table=0, ...
    #  cookie=0xb7dff131a697c6a5, duration=2411726.786s, table=23, ...
    #  cookie=0xb7dff131a697c6a5, duration=2411726.760s, table=24, ...
    #
    return 'NXST' not in line and 'OFPST' not in line



# Copyright 2016 Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.objects import flavor
from neutron.tests.unit.objects import test_base as obj_test_base
from neutron.tests.unit import testlib_api


class FlavorServiceProfileBindingIfaceObjectTestCase(
        obj_test_base.BaseObjectIfaceTestCase):

    _test_class = flavor.FlavorServiceProfileBinding


class FlavorServiceProfileBindingDbObjectTestCase(
        obj_test_base.BaseDbObjectTestCase, testlib_api.SqlTestCase):

    _test_class = flavor.FlavorServiceProfileBinding

    def setUp(self):
        super(FlavorServiceProfileBindingDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'flavor_id': lambda: self._create_test_flavor_id(),
             'service_profile_id':
                 lambda: self._create_test_service_profile_id()})


class ServiceProfileIfaceObjectTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = flavor.ServiceProfile


class ServiceProfileDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                                     testlib_api.SqlTestCase):

    _test_class = flavor.ServiceProfile

    def test_get_objects_queries_constant(self):
        # FIXME(electrocucaracha): There are no lazy loading for flavors
        # relationship in ServiceProfile model db disable this UT to avoid
        # failing
        pass


class FlavorIfaceObjectTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = flavor.Flavor


class FlavorDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                             testlib_api.SqlTestCase):

    _test_class = flavor.Flavor

    def test_get_objects_queries_constant(self):
        # FIXME(electrocucaracha): There are no lazy loading for
        # service_profiles relationship in Flavor model db disable this UT to
        # avoid failing
        pass
